
from dreem.util.util import *
import pyarrow.orc as po
import pyarrow as pa

def generate_mut_profile_from_bit_vector(bit_vector, clustering_file, verbose=False):
    """
    Generate a mutation profile from a bit vector.

    Parameters
    ----------
    bit_vector : str
        Path to the bit vector.
    verbose : bool
        If True, print progress.

    Returns
    -------
    df_row: dict
        - info_bases: the count per residue of bases for which we have information.
        - cov_bases: the count per residue of bases for which we have coverage.
        - mut_bases: the count per residue of bases for which we have mutations.
        - mod_bases_A: the count per residue of bases who are modified to A.
        - mod_bases_C: the count per residue of bases who are modified to C.
        - mod_bases_G: the count per residue of bases who are modified to G.
        - mod_bases_T: the count per residue of bases who are modified to T.
        - mut_rates: the mutation rate per residue.
        - num_aligned: the number of aligned reads.
        
    out['sequence'] = ''.join([c[-1] for c in bv.column_names])
    out['num_aligned'] = muts.shape[0]
    out["match_bases"] = query_muts(muts, MATCH[0])
    out["mod_bases_A"] = query_muts(muts, SUB_A[0])
    out["mod_bases_C"] = query_muts(muts, SUB_C[0])
    out["mod_bases_G"] = query_muts(muts, SUB_G[0])
    out["mod_bases_T"] = query_muts(muts, SUB_T[0])
    out["mod_bases_N"] = query_muts(muts, SUB_N[0])
    out["del_bases"]   = query_muts(muts, DELET[0])
    out["ins_bases"]   = query_muts(muts, INS_5[0] | INS_3[0])
    # Can have any mutation, but not a match
    out["mut_bases"] = query_muts(muts, SUB_N[0] | DELET[0] | INS_5[0] | INS_3[0])
    out["cov_bases"] = muts.astype(bool).sum(axis=0)  # i.e. not BLANK
    # Unambiguously matching or mutated (informative)
    out["info_bases"] = out["match_bases"] + out["mut_bases"]
    # Mutation rate (fraction mutated among all unambiguously matching/mutated)
    try:
        out["mut_rates"] = out["mut_bases"] / out["info_bases"]
    except ZeroDivisionError:
        out["mut_rates"] = [m/i if i != 0 else None for m, i in zip(out["mut_bases"], out["info_bases"])]
    
    out['num_of_mutations'] =  query_muts(muts, SUB_N[0] | DELET[0] | INS_5[0] | INS_3[0], axis=1)

    """
    # Read in the bit vector
    bv = pa.concat_tables([po.read_table(os.path.join(bit_vector,b)) for b in os.listdir(bit_vector) if b.endswith(".orc")])
    muts = np.array(bv, dtype=np.uint8).T
    # Convert to a mutation profile
    out = dict()
    out['sequence'] = ''.join([c[0] for c in bv.column_names])
    out['num_aligned'] = muts.shape[0]
    out["match_bases"] = query_muts(muts, MATCH[0] | INS_5[0], set_type='subset')
    out["mod_bases_A"] = query_muts(muts, SUB_A[0], set_type='subset')
    out["mod_bases_C"] = query_muts(muts, SUB_C[0], set_type='subset')
    out["mod_bases_G"] = query_muts(muts, SUB_G[0], set_type='subset')
    out["mod_bases_T"] = query_muts(muts, SUB_T[0], set_type='subset')
    out["mod_bases_N"] = query_muts(muts, SUB_N[0], set_type='subset')
    out["del_bases"]   = query_muts(muts, DELET[0], set_type='subset')
    out["ins_bases"]   = query_muts(muts, INS_3[0], set_type='superset')
    # Can have any mutation, but not a match
    out["mut_bases"] = out["mod_bases_N"] #query_muts(muts, SUB_N[0] | DELET[0] | INS_3[0], set_type='superset')
    out["cov_bases"] = muts.astype(bool).sum(axis=0)  # i.e. not BLANK
    # Unambiguously matching or mutated (informative)
    out["info_bases"] = out["match_bases"] + out["mut_bases"]
    # Mutation rate (fraction mutated among all unambiguously matching/mutated)
    try:
        out["mut_rates"] = out["mut_bases"] / out["info_bases"]
    except ZeroDivisionError:
        out["mut_rates"] = [m/i if i != 0 else None for m, i in zip(out["mut_bases"], out["info_bases"])]
    
    out['num_of_mutations'] = np.histogram(query_muts(muts, SUB_N[0], axis=1), bins=range(0, muts.shape[1]))[0] # query_muts(muts, SUB_N[0] | DELET[0] | INS_3[0], axis=1)
    
    out['worst_cov_bases'] = min(out['cov_bases'])
    
    out.pop("match_bases")
    for k in out:
        if isinstance(out[k], np.ndarray):
            out[k] = list(out[k])
    return out




import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import pytest
import pandas as pd
import dreem.util as util
import os
from dreem import clustering
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import json 
import numpy as np

module = 'clustering'

sample_name = 'test_set_1'
MAX_CLUSTERS = 2

reads_partition = [100000]*2

half_sequence_length = 100
unpaired_bases = int(0.3*half_sequence_length)
shared_bases = [int(u*unpaired_bases) for u in np.linspace(0.0, 0.8, 4)]
mu_unpaired = [0.07]

sample_profile = {}
for sc in shared_bases:
    for mu in mu_unpaired:
        profile_name = 'r{}_ub{}_sb{}_mu{}'.format(reads_partition, unpaired_bases, sc, int(mu*100))
        sample_profile[profile_name] = {
            'n_reads': reads_partition,
            'n_AC': half_sequence_length,
            'n_unpaired': unpaired_bases,
            'n_shared': sc,
            'path_bv': os.path.join(test_files_dir, 'input', module, sample_name, 'r{}_ub{}_sb{}_mu{}/0.orc'.format(reads_partition, unpaired_bases, sc, int(mu*100))),
            'path_json': os.path.join(test_files_dir, 'output', module, sample_name, profile_name),
            'mu_unpaired': [mu]*2
        }
    
module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

inputs = ['clustering']

@pytest.mark.skip(reason="Too bugged")
def test_make_files():
    os.makedirs(os.path.join(test_files_dir, 'input', module, sample_name), exist_ok=True)
    os.makedirs(os.path.join(test_files_dir, 'expected_output', module, sample_name), exist_ok=True)
    files_generator.generate_files(sample_profile, module, inputs, [], test_files_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    
@pytest.mark.skip(reason="Too bugged")
def test_run():
    clustering.run(
        input_dir = os.path.join(module_input, sample_name),
        out_dir = module_output,
        num_runs= 5,
        min_iter=30,
        max_clusters=MAX_CLUSTERS,
        n_cpus=10
    )

@pytest.mark.skip(reason="Too bugged")
def test_assert_files_exist():
    assert os.path.exists(os.path.join(module_output, 'best_cluster_reads.json'))

@pytest.mark.skip(reason="not implemented")
def test_assess_performance():

    for name, sample in sample_profile.items():
    
        with open(os.path.join(module_output, 'best_cluster_reads.json'), 'r') as f:
            result_dict = json.load(f)
            n_reads = [ sum([k[1]==str(idx+1) for k in result_dict[name]['K2'].keys()]) for idx in range(len(sample['n_reads']))]

            result_np = pd.DataFrame.from_dict(result_dict[name]['K'+str(len(sample['n_reads']))]).to_numpy().T
            F1 = 1.0

            if len(n_reads) == 2:
                F1_class = []
                for i in range(2):
                    TP = np.count_nonzero(np.argmax(result_np[:n_reads[0]], axis=1)==i)
                    if TP != 0:
                        precision = TP/np.count_nonzero(np.argmax(result_np, axis=1)==i)
                        recall = TP/n_reads[0]
                        F1_class.append(2*precision*recall/(precision+recall))
                F1 = max(F1_class)
    
            if len(n_reads) == 3:
                n_reads_cumulative = [0] + n_reads[:-1]
                F1 = 0
                indices = [0,1,2]
                
                # Compute F1 score of each class for all combinations
                for c in range(len(n_reads)):
                    F1_class = []
                    start = sum(n_reads_cumulative[:c+1])
                    end = start + n_reads[c]
                    for i in indices:
                        TP = np.count_nonzero(np.argmax(result_np[start:end], axis=1)==i)
                        if TP != 0:
                            precision = TP/np.count_nonzero(np.argmax(result_np, axis=1)==i)
                            recall = TP/n_reads[c]
                            F1_class.append(2*precision*recall/(precision+recall))
                        else:
                            F1_class.append(0.0)
                    indices.pop(np.argmax(F1_class))
                    F1 += np.max(F1_class)
                F1 /= 3

            print('F1 score: ', F1)
            sample['F1'] = F1
            assert F1 > 0.5

                

if __name__ == '__main__':
    test_make_files()
    test_run()
    test_assert_files_exist()
    test_assess_performance()



from dreem.util.seq import parse_fasta
import numpy as np
from dreem.aggregate.resources.get_attributes import read_sample_attributes
import pandas as pd
import os

def check_library(library: pd.DataFrame, path_fasta: str, path_save_clean_library: str = None):
    """Sanity check for library.csv
    
    # sections
    - Check that there is only one value for each construct for each column that's not 'construct', 'section', 'section_start', 'section_end'
    - Check that every construct is in the fasta file
    - Check that there are no duplicate rows
    - If there are multiple sections, copy paste the other attributes so that every row is covered
    - Check that the section_start and section_end are in the correct order
    - Check that the section_start and section_end are within the length of the corresponding sequence in the fasta file.
    - If there are multiple sections, check that the section_start and section_end are unique. If not, remove the duplicates.
    - If section is empty, fill it with the section_start and section_end values separated by an underscore
    - If a construct has no sections, name the section 'full'
    
    # barcodes
    - Check that every barcode is in the fasta file at the given position for the given construct
    - Check that the barcode_start is within the length of the corresponding sequence in the fasta file.
    - Check that the barcode_start is unique for each construct
    - Check that each barcode is unique for each construct
    
    Args:
        library (pandas.DataFrame): content of library.csv
        path_fasta (str): path to the fasta file
        path_save_clean_library (str, optional): path to save the clean library.csv. Defaults to None.
        
    Returns:
        Library (pandas.DataFrame): reframed content of library.csv
    
    """
    

    a = parse_fasta(path_fasta)
    fasta = pd.DataFrame({k.decode("utf-8") :v.decode("utf-8")  for k,v in a}, index=[0]).T.reset_index().rename(columns={"index":"construct", 0:"sequence"})

    # Check that the constructs in the library are in the fasta file
    if not set(library["construct"].unique()).issubset(set(fasta["construct"].unique())):
        raise ValueError("Some constructs are not in the fasta file")

    # Make sure that barcode_start, section_start and section_end are integers
    cols = []
    cols += ["barcode_start"] if "barcode_start" in library.columns else []
    cols += ["section_start"] if "section_start" in library.columns else []
    cols += ["section_end"] if "section_end" in library.columns else []
    for col in cols:
        library[col] = library[col].apply(lambda x: int(x) if not pd.isnull(x) else x).astype("Int64")
    
    # Sections
    # Check that there is only one value for each construct for each column that's not 'construct', 'section', 'section_start', 'section_end'
    if 'section' in library.columns:
        for idx, g in library.groupby("construct"):
            for col in g.columns:
                if col not in ["construct", "section", "section_start", "section_end"]:
                    un = g[col].unique()
                    if np.count_nonzero(~pd.isnull(un)) > 1:
                        raise ValueError(f"{col} has multiple values for {idx}")
        
        # Check that every construct is in the fasta file
        if not set(library["construct"].unique()).issubset(set(fasta["construct"].unique())):
            raise ValueError("Some constructs are not in the fasta file")
        
        # Check that there are no duplicate rows and remove them
        if library.duplicated().any():
            library = library.drop_duplicates()
            
        # If there are multiple sections, copy paste the other attributes so that every row is covered
        if library["section"].nunique() > 1:
            for idx, g in library.groupby("construct"):
                for col in g.columns:
                    if col not in ["construct", "section", "section_start", "section_end"]:
                        library.loc[library["construct"] == idx, col] = g[col].unique()[0]
    

    if 'section' in library.columns:     
        # Check that the section_start and section_end are in the correct order
        if (library["section_start"] > library["section_end"]).any():
            raise ValueError("section_start is greater than section_end")
        
        # Check that the section_start is 1 or greater
        if (library["section_start"] < 1).any():
            raise ValueError("section_start is less than 1")
        
        # Check that section_end is within the length of the corresponding sequence in the fasta file.
        for idx, g in library.groupby("construct"):
            sequence = fasta.loc[fasta["construct"] == idx, "sequence"].unique()[0]
            if (g["section_end"] > len(sequence)).any():
                raise ValueError(f"section_end is greater than the length of the sequence for {idx} and section {max(g['section'].unique())} (sequence length: {len(sequence)})")
            
        # If there are multiple sections, check that the section_start and section_end are unique. If not, remove the duplicates.
        if library["section"].nunique() > 1:
            if library.duplicated(subset=["construct", "section_start", "section_end"]).any():
                library = library.drop_duplicates(subset=["construct", "section_start", "section_end"])
        
        # Assert no section is named 'full'
#        if "full" in library["section"].unique():
#            raise ValueError("Section cannot be named 'full'")
        
        # If section, section_start, and section_end are all empty for a certain row, fill section of this row with 'full' and set the section_start and section_end to 0 and the length of the sequence
        for idx, row in library.iterrows():
            if np.count_nonzero(pd.isnull(row[['section', 'section_start', 'section_end']])) == 3:
                library.loc[idx, "section_start"] = 1
                library.loc[idx, "section_end"] = len(fasta.loc[fasta["construct"] == row["construct"], "sequence"].unique()[0])
                library.loc[idx, "section"] = '1-'+str(library.loc[idx, "section_end"])
        
        # If section is empty but not the section start and end, fill it with the section_start and section_end values separated by an underscore
        if library["section"].isna().any():
            library.loc[library["section"].isna(), "section"] = library.loc[library["section"].isna(), "section_start"].astype(str) + "-" + library.loc[library["section"].isna(), "section_end"].astype(str)
        
    # Barcodes
    if "barcode" in library.columns:
        # Check that the barcode_start is unique for each construct
        if library.groupby("construct")["barcode_start"].nunique().max() > 1:
            raise ValueError("barcode_start is not unique for each construct")
            
        # Check that the barcode_start is within the length of the corresponding sequence in the fasta file.
        for idx, g in library.groupby("construct"):
            sequence = fasta[fasta["construct"] == idx]["sequence"].unique()[0]
            if (g["barcode_start"] + len(g['barcode'])> len(sequence)).any():
                raise ValueError(f"barcode_start + lenght of barcode is greater than the length of the sequence for {idx}")

        # Check that every barcode is in the fasta file at the given position for the given construct
        for idx, g in library.groupby("construct"):
            barcode = g["barcode"].unique()[0]
            barcode_start = g["barcode_start"].unique()[0] - 1
            if barcode != fasta.loc[fasta["construct"] == idx, "sequence"].unique()[0][barcode_start:barcode_start+len(barcode)]:
                raise ValueError(f"Barcode {barcode} is not in the fasta file at the given position for {idx}")
    
    # Check taht each attribute is unique for each construct
    for col in library.columns:
        if not col.startswith("section") and not col.startswith("barcode") and col not in ["construct"]:
            if library.groupby("construct")[col].nunique().max() > 1:
                raise ValueError(f"{col} is not unique for each construct")
            for construct in library.construct.unique():
                library.loc[library["construct"] == construct, col] = library.loc[library["construct"] == construct, col].unique()[0]
    
    if path_save_clean_library is not None and os.path.exists(path_save_clean_library): #
        library.to_csv(os.path.join(path_save_clean_library,'clean_library.csv'), index=False)
    return library

def check_samples(samples):
    """Check that the samples csv file is correct
    
    - Check that the columns are correct
    - Check that "exp_env" is either "in_vivo" or "in_vitro"
    - Check that the sample names are unique
    - Check that no mandatory columns are missing or empty
    
    Args:
        samples (pandas.DataFrame): The samples csv file
        
    Returns:
        pandas.DataFrame: The samples csv file
    """
    
    samples_attributes = read_sample_attributes()['mandatory']
    mandatory_columns = samples_attributes['all']
    in_vivo_cols = samples_attributes['in_vivo']
    in_vitro_cols = samples_attributes['in_vitro']
    
    # Check that the columns are correct
    if not set(mandatory_columns).issubset(set(samples.columns)):
        missing_col = list(set(mandatory_columns) - set(samples.columns))
        raise ValueError(f"Some mandatory columns are missing. Columns are: {missing_col}")

    if 'in_vivo' in samples['exp_env'].unique():
        if not set(in_vivo_cols).issubset(set(samples.columns)):
            missing_col = list(set(in_vivo_cols + mandatory_columns) - set(samples.columns))
            raise ValueError(f"Some mandatory columns are missing. Columns are: {missing_col}")
        
    if 'in_vitro' in samples['exp_env'].unique():
        if not set(in_vitro_cols).issubset(set(samples.columns)):
            missing_col = list(set(in_vitro_cols + mandatory_columns) - set(samples.columns))
            raise ValueError(f"Some mandatory columns are missing. Columns are: {missing_col}")
    
    # Check that "exp_env" is either "in_vivo" or "in_vitro"
    if not set(samples['exp_env'].unique()).issubset(set(['in_vivo', 'in_vitro'])):
        raise ValueError("exp_env can only be 'in_vivo' or 'in_vitro'")
    
    for _, sample in samples.iterrows():
        # Check that the sample names are unique
        if samples[samples['sample'] == sample['sample']].shape[0] > 1:
            raise ValueError(f"Sample name {sample['sample']} is not unique")
        
        # Check that no mandatory columns are missing or empty
        if sample['exp_env'] == 'in_vivo':
            if sample[in_vivo_cols].isna().any():
                raise ValueError(f"Sample {sample['sample']} has missing or empty mandatory columns")
        elif sample['exp_env'] == 'in_vitro':
            if sample[in_vitro_cols].isna().any():
                raise ValueError(f"Sample {sample['sample']} has missing or empty mandatory columns")
        else:
            raise ValueError(f"Sample {sample['sample']} has an invalid sample_type")
    
    return samples
    
    
def compare_fields(d1, d2, fields):
    assert read_dict(d1, fields) == read_dict(d2, fields), 'expected and real output are not equal for field {}'.format('/'.join(fields))
    
def read_dict(d, fields):
    if fields != []:
        if fields[0] not in d:
            raise ValueError('Key {} not found. Check if it\'s implemented.'.format(fields[0]))
        return read_dict(d[fields[0]], fields[1:])
    return d


import click
from main import run
from click_option_group import optgroup
from dreem.util.cli import *


@click.command()
@opti_library
@opti_coords
@opti_primers
@opti_fill
@opti_parallel
@opto_out_dir
@argi_fasta
@argi_bams
def cli(*args, **opts):
    run(*args, **opts)


if __name__ == '__main__':
    cli()




import numpy as np
import pyarrow as pa
import pyarrow.orc as po
import pyarrow.compute as pc

from dreem.util.util import *

class BitVector:
    """Container object. Contains the name of the construct, the sequence, the bitvector, the read names and the read count.
    """
    
    def __init__(self,path, **args) -> None:
        preprocessing = self.preprocessing(path)

        self.name = path.split('/')[-1][:-(len('.orc'))]
        self.sequence = preprocessing[0]
        self.bv = preprocessing[1]
        self.read_index = preprocessing[2]
        self.read_inverse = preprocessing[3]
        self.read_hist = preprocessing[4]
        self.read_names = preprocessing[5]
        self.report = preprocessing[6]
        self.base_to_keep = preprocessing[7]
        self.publish_preprocessing_report(path=path[:-len('.orc')]+'_preprocessing_report.txt')
        
    #TODO optimize this 
    def preprocessing(self, path, low_mut_rate = 0.015, use_G_U = False, max_mut_close_by = 4):
        """Preprocess the bitvector.
        
        - Remove the bases G and U
        - Count and remove duplicates
        - Remove the bases with a mutation rate below low_mut_rate
        - save the read names and the read count
        
        Parameters:
        -----------
        
        path: str
            Path to the directory containing the bitvector.
            
        low_mut_rate: float
            Mutation rate below which the bases are removed.
        
        use_G_U: bool
            If True, keep the bases G and U.
            
            
        Output:
        -------
        
        bv: numpy array (N x D)
            Preprocessed bitvector.
            
        read_hist: numpy array (N)
            Count of reads per bitvector.
            
        read_names: list of str
            Read names.
            
        report: str
            Report of the preprocessing.
            #TODO define
            
        """
        
        report = {}
        
        bv = po.read_table(path)
        report['total_number_of_reads'] = bv.shape[0]     
        
        # Take the read names
        read_names = np.array(bv.column('id'), dtype = str)
        bv = bv.drop(['id'])   
        
        ## PER BASE REMOVALS
        # Remove the non-informative bases types
        sequence = ''.join([c[0] for c in bv.column_names])  
        report['sequence'] = sequence
        
        temp_n_cols = bv.shape[1]
        bases_to_drop = set()
        if not use_G_U:
            # bv = bv.drop([c for c in bv.column_names if c[0] in ['G','T']])
            bases_to_drop |= set([c for c in bv.column_names if c[0] in ['G','T']])
        report['removed_G_U'] = temp_n_cols - bv.shape[1]

        # Remove the low-mutation-rate bases
        bin_bv = mutations_bin_arr(bv)
        mut_rates = np.sum(bin_bv, axis = 1)/bin_bv.shape[1]
        paired_bases = [unpaired for unpaired, mut_rate in zip(bv.column_names, mut_rates) if mut_rate < low_mut_rate]
        bases_to_drop |= set(paired_bases)
        bases_to_keep = set(bv.column_names)-bases_to_drop
        bases_to_keep = [int(i[1:]) for i in bases_to_keep]; bases_to_keep.sort()

        temp_n_cols = bv.shape[1]
        bv = bv.drop(bases_to_drop)
        report['too_low_mutation_rate'] = temp_n_cols - bv.shape[1]
        report['min_mutation_rate'] = low_mut_rate


        ## PER READ REMOVALS
        # Remove the bit vectors with too many mutations
        temp_n_reads = bv.shape[0]
        bin_bv = mutations_bin_arr(bv)
        n_muts_per_reads = np.sum(bin_bv, axis = 0)
        idx = np.nonzero( n_muts_per_reads < np.mean(n_muts_per_reads) + 3*np.std(n_muts_per_reads) )[0]
        bv, read_names = pc.take(bv, idx), read_names[idx]
        report['too_many_mutations'] = temp_n_reads - bv.shape[0]
        
        # Remove the bitvectors with deletions
        temp_n_reads = bv.shape[0]
        dels = deletions_bin_array(bv)
        no_deletion_in_the_read = np.nonzero(~np.sum(dels, axis=0, dtype=bool))[0]
        # MAKE BV BINARY
        for i, c in enumerate(bv.column_names):
            bv = bv.set_column(i, c, [mutations_bin_arr(bv.column(c))])
        bv, read_names = pc.take(bv, no_deletion_in_the_read), read_names[no_deletion_in_the_read]
        report['no_info_around_mutations'] = temp_n_reads - bv.shape[0]

        #Remove the bit vectors with 'max_mut_close_by' consecutive mutations
        idx_remove_consecutive_mutations = []
        for i0, c0 in enumerate(bv.column_names[:-1]):
            for c1 in bv.column_names[i0+1:i0+max_mut_close_by+1]:
                if int(c1[1:]) - int(c0[1:]) <= max_mut_close_by:
                    idx_remove_consecutive_mutations += np.nonzero(np.logical_and(bv[c0], bv[c1]))[0].tolist()
        mask = np.ones(bv.shape[0], dtype=bool)
        mask[idx_remove_consecutive_mutations] = False
        temp_n_reads = bv.shape[0]
        bv, read_names = pc.take(bv, np.arange(bv.shape[0])[mask]), read_names.take(np.arange(bv.shape[0])[mask])

        # Turn bv into a np array
        bv = np.array(bv, dtype = np.uint8).T        
        report['mutations_close_by'] = temp_n_reads - bv.shape[0] 
        
            
        # What's this #TODO
        report['too_few_informative_bits'] = '#TODO'
        
        # Remove the duplicates and count the reads
        bv, read_idx, read_inverse, read_hist = np.unique(bv, axis = 0, return_index=True, return_inverse=True, return_counts = True)
        report['number_of_unique_reads'] = read_hist.shape[0]
        report['number_of_used_reads'] = np.sum(read_hist)
        report['bases_used'] = bv.shape[1]
        
        
        # Sanity check
        assert len(report['sequence']) == report['bases_used'] + report['too_low_mutation_rate'] + report['removed_G_U']
        assert report['total_number_of_reads'] == report['number_of_used_reads'] + report['too_many_mutations'] + report['no_info_around_mutations'] + report['mutations_close_by']
        return sequence, bv, read_idx, read_inverse, read_hist, read_names, report, bases_to_keep
    

   
   
    def publish_preprocessing_report(self, path):
        """Publish the report in a text file.

        Parameters:
        -----------

        path: str
            Path to the directory where the report will be saved.
            
        """

        report = "Sequence: "+ self.report['sequence'] \
        + "\nLength of the sequence: " + str(len(self.report['sequence'])) \
        + "\nResidues used: " + str(self.report['bases_used']) \
        + "\nResidues removed because they are Gs and Us: " + str(self.report['removed_G_U']) \
        + "\nResidues removed because they have mutation rate below {} : {}".format(self.report['min_mutation_rate'], self.report['too_low_mutation_rate']) \
        + "\nNumber of bit vectors in total: "+ str(self.report['total_number_of_reads'])\
        + "\nNumber of bit vectors used: "+ str(self.report['number_of_used_reads']) \
        + "\nNumber of unique bit vectors used: " + str(self.report['number_of_unique_reads']) \
        + "\nNumber of bit vectors discarded: "+ str(self.report['total_number_of_reads'] - self.report['number_of_used_reads']) \
        + "\nBit vectors removed because of too many mutations: " + str(self.report['too_many_mutations']) \
        + "\nBit vectors removed because of too few informative bits: " + str(self.report['too_few_informative_bits']) \
        + "\nBit vectors removed because of mutations close by: " + str(self.report['mutations_close_by']) \
        + "\nBit vectors removed because of no info around mutations:  " + str(self.report['no_info_around_mutations']) 
        with open(path, 'w') as f:
            print(report)
            f.write(report)
           
           
    def associate_reads_with_likelihoods(self, likelihood_per_read):
        """Associates the reads with their likelihood, using the attributes read_names and read_hist. 
        Publish the reads in a json file.

        Parameters:
        -----------

        likelihood_per_read: array (N x K)
            Likelihood of each read.

        Output:
        -------

        reads: dict
            Dictionary associating the read name with the likelihood.

        """

        reads = {}
        for name, idx in zip(self.read_names, self.read_inverse):
            reads[name] = {}
            for k in range(likelihood_per_read.shape[1]):
                reads[name]['K'+str(k+1)] = likelihood_per_read[idx,k]

        return reads


def mutations_bin_arr(bv):
    return query_muts(np.array(bv, dtype=np.uint8), SUB_N[0] | DELET[0] | INS_5[0] | INS_3[0], sum_up=False)


def deletions_bin_array(bv):
    return query_muts(np.array(bv, dtype=np.uint8), DELET[0], sum_up=False)


import pandas as pd
import numpy as np

from  dreem.draw.util import *

import plotly.graph_objects as go
from plotly.offline import plot, iplot

from dreem.draw import  util
from itertools import cycle
from typing import Tuple, List
from sklearn import metrics
from sklearn.linear_model import LogisticRegression


LIST_COLORS = ['red','green','blue','orange','purple','black','yellow','pink','brown','grey','cyan','magenta']

def mutation_fraction(df, show_ci:bool=True, savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    assert len(df) == 1, "df must have only one row"
    mh = df.iloc[0]
    cmap = {"A": "red", "T": "green", "G": "orange", "C": "blue"}  # Color map
    
    traces, layouts = [], []
    mh_unrolled = pd.DataFrame({'mut_rate':list(mh.mut_rates), 'base':list(mh.sequence), 'index_reset':list(range(len(mh.index_selected))),'index_selected':list(mh.index_selected), 'poisson_high':list(mh.poisson_high), 'poisson_low':list(mh.poisson_low), 'paired':list(mh.structure_selected)})

    for bt in set(mh['sequence']):
        df_loc = mh_unrolled[mh_unrolled['base'] == bt]
        if len(df_loc) == 0:
            continue

        hover_attr = pd.DataFrame({'mut_rate':list(df_loc.mut_rate),
                                        'base':list(df_loc.base), 
                                        'index':list(df_loc['index_selected']),
                                        'paired':[{'.':True, '(':False,')':False}[s] for s in df_loc.paired]})
        traces.append(go.Bar(
            x= np.array(df_loc['index_reset']),
            y= np.array(df_loc['mut_rate']),
            name=bt,
            marker_color=cmap[bt],
            text = hover_attr,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            ))
        if show_ci:
            traces[-1].update(
                        error_y=dict(
                        type='data',
                        symmetric=False,
                        array=df_loc['poisson_high'], 
                        arrayminus=df_loc['poisson_low']
                        ))

    
        mut_fig_layout = go.Layout(

        )

    fig = go.Figure(data=traces, layout=mut_fig_layout)

    fig.update_layout(title=f"{mh['sample']} - {mh['construct']} - {mh['section']} - {mh['cluster']} - {mh['num_aligned']} reads",
                        xaxis=dict(title="Sequence"),
                        yaxis=dict(title="Mutation rate", range=[0, 0.1]))
   
    fig.update_yaxes(
            gridcolor='lightgray',
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )
    fig.update_xaxes(
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )

    fig.update_xaxes(
            tickvals=mh_unrolled['index_reset'],
            ticktext=["%s %s" % ({'.':'(P)','(':'(U)',')':'(U)'}[x], str(y)) for (x,y) in zip(mh['structure_selected'],mh['index_selected'])],
            tickangle=90,
            autorange=True
    )

    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)

    return {'fig':fig, 'df':mh}


def deltaG_vs_mut_rates(df:pd.DataFrame, models:List[str]=[],  savefile=None, auto_open=False, use_iplot=True, title=None)->dict:

    df_temp = pd.DataFrame()
    for _, row in df.iterrows():
        df_temp = pd.concat([df_temp, pd.DataFrame({'construct':row.construct, 'index':row.index_selected, 'mut_rates':row.mut_rates, 'num_aligned':row.num_aligned, 'deltaG':row['deltaG_selected'],'base':list(row.sequence), 'paired':[s !='.' for s in row.structure_selected]}, index=list(range(len(row.index_selected))))])
    
    assert len(df_temp) > 0, "No data to plot"
    df = df_temp.reset_index()

    hover_attr = ['num_aligned','mut_rates','base','index','construct','deltaG']
    tra = {}
    for is_paired, prefix in zip([True,False], ['Paired ','Unpaired ']):
        markers = cycle(list(range(153)))
        x=np.array(df[df.paired == is_paired]['deltaG'])
        y=np.array(df[df.paired == is_paired]['mut_rates'])
        tra[prefix] = go.Scatter(
            x=x,
            y=y,
            text = df[df.paired == is_paired][hover_attr],
            marker_size= df[df.paired == is_paired]['num_aligned']/200,
            mode='markers',
            name=prefix,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            line=dict(color='green' if is_paired else 'red'))
            
        
        for m in models:
            if len(y) > 0:
                fit = util.Fit()
                x_sorted, pred_y_sorted = fit.predict(x,y,m, prefix)
                tra[fit.get_legend()] = go.Scatter(
                    x=x_sorted,
                    y=pred_y_sorted,
                    mode='lines+markers',
                    name=fit.get_legend(), 
                    marker=dict(symbol=next(markers)),
                    line=dict(color='darkseagreen' if is_paired else 'crimson', dash='dash'))

    layout = dict(title = 'Mutation rates of paired / unpaired residues vs the expected energy of the molecule',
            xaxis= dict(title= 'DeltaG',ticklen= 5,zeroline= False),
            yaxis= dict(title= 'Mutation rate ',ticklen= 5,zeroline= False),
            )

    fig = dict(data = list(tra.values()), layout = layout)

    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)


    return {'fig':fig, 'df':df}
    
def exp_variable_across_samples(df:pd.DataFrame, experimental_variable:str, models:List[str]=[],  savefile=None, auto_open=False, use_iplot=True, title=None)->dict:

    colors = cycle(LIST_COLORS)
    data = pd.DataFrame()
    for _, row in df.iterrows():
        data = pd.concat([data, pd.DataFrame({'sample':row['sample'],experimental_variable:getattr(row,experimental_variable), 'index':list(row.index_selected), 'base':list(row.sequence), 'mut_rates':list(row.mut_rates), 'paired':[s !='.' for s in row.structure_selected]}, index=list(range(len(row.index_selected))))])
    data = data.reset_index().rename(columns={'level_0':'index_subsequence'})
    data = data.sort_values(by='index')
    data['Marker'] = data['paired'].apply(lambda x: {True: 0, False:1}[x])
    hover_attr = ['base','mut_rates','sample',experimental_variable, 'paired']

    tra = {}
    for idx, row in data.groupby('index_subsequence'):
        color = next(colors)
        markers = cycle(list(range(153)))
        name = f"({row['base'].iloc[0]},{row['index'].iloc[0]})"
        tra[row['index'].iloc[0]] = go.Scatter(
            x= row[experimental_variable], 
            y= row['mut_rates'], 
            text = data[hover_attr],
            mode='lines+markers',
            marker = dict(symbol = list(map(util.Setshape, data['Marker']))),
            name= name,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            line=dict(color=color))

        my_dash = cycle(['solid', 'dot', 'dash', 'longdash', 'dashdot', 'longdashdot'])
        for m in models:
            x= row[experimental_variable]
            y= row['mut_rates']
            fit = util.Fit()
            x_sorted, pred_y_sorted = fit.predict(x,y,m, name)
            tra[fit.get_legend()] = go.Scatter(
                x=x_sorted,
                y=pred_y_sorted,
                mode='lines',
                line=dict(dash= next(my_dash), color=color),
               # line=dict(color=color),
                name=fit.get_legend())

    layout = dict(title = 'Mutation rates of paired / unpaired residues vs '+experimental_variable,
            xaxis= dict(title= experimental_variable,ticklen= 5,zeroline= False),
            yaxis= dict(title= 'Mutation rate ',ticklen= 5,zeroline= False),
            )

    #tra = {arg:tra[list(tra.keys()[arg])] for arg in np.argsort(np.array([int(k[k.index('(')+3:k.index(')')]) for k in tra.keys()]))}

    fig = dict(data = list(tra.values()), layout = layout)
    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)

    return {'fig':fig, 'df':data}



def auc(df:pd.DataFrame,  savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    def make_roc_curve(X, y, y_pred, fig, title):
        fpr, tpr, thresholds = metrics.roc_curve(y, y_pred)
        roc_auc = metrics.auc(fpr, tpr)
        fig.add_trace(go.Scatter(x=fpr, y=tpr,
                            mode='lines',
                            name=title,
                            line=dict(width=2)))
        return fig


    fig = go.Figure()
    for row in df.iterrows():
        X = row[1]['mut_rates'].reshape(-1, 1)
        y = np.array([1 if c == '.' else 0 for c in row[1]['structure_selected']]).reshape(-1, 1)
        y_pred = LogisticRegression().fit(X, y.ravel()).predict_proba(X)[:,1]
        fig = make_roc_curve(X, y, y_pred, fig, row[1]['unique_id'])


    fig.add_trace(
        go.Scatter(x=[0, 1], y=[0, 1], line=dict(color='black', width=2, dash='dash'), showlegend=False)
    )
    fig.update_layout(
        title='ROC Curve',
        xaxis_title='False Positive Rate',
        yaxis_title='True Positive Rate')

    fig.update_yaxes(scaleanchor="x", scaleratio=1)
    fig.update_xaxes(constrain='domain')

    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)

    return {'fig':fig, 'df':df}


def mutation_fraction_delta(df, savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    assert len(df) == 2, "df must have 2 row"
    mp_attr = ['sample', 'construct', 'section', 'cluster']
    df['unique_id'] = df.apply(lambda row: ' - '.join([str(row[attr]) for attr in mp_attr]), axis=1)

    mh = pd.Series(
        {
            'mut_rates': df['mut_rates'].values[0] - df['mut_rates'].values[1],
            'sequence': ''.join([c1 if c1 == c2 else '-' for c1,c2 in zip(df['sequence'].values[0],df['sequence'].values[1])]),
            'title': "{} - {} reads vs {} - {} reads".format(df['unique_id'].values[0], df['num_aligned'].values[0], df['unique_id'].values[1], df['num_aligned'].values[1])
        }
    )
    cmap = {"A": "red", "T": "green", "G": "orange", "C": "blue", '-':'grey'}  # Color map
    
    traces, layouts = [], []
    mh_unrolled = pd.DataFrame({'mut_rate':list(mh.mut_rates), 'base':list(mh.sequence), 'index_reset':list(range(len(mh.sequence)))})

    for bt in set(mh['sequence']):
        df_loc = mh_unrolled[mh_unrolled['base'] == bt]
        if len(df_loc) == 0:
            continue

        hover_attr = pd.DataFrame({'mut_rate':list(df_loc.mut_rate),
                                        'base':list(df_loc.base)})
        traces.append(go.Bar(
            x= np.array(df_loc['index_reset']),
            y= np.array(df_loc['mut_rate']),
            name=bt,
            marker_color=cmap[bt],
            text = hover_attr,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            ))
    

    fig = go.Figure(data=traces, 
                    layout=go.Layout(
                        title=go.layout.Title(text=mh['title']),
                        xaxis=dict(title="Sequence"),
                        yaxis=dict(title="Mutation rate", range=[0, 0.1])))

    fig.update_yaxes(
            gridcolor='lightgray',
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )
    fig.update_xaxes(
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )

    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)
    
    return {'fig':fig, 'df':mh}
 

def base_coverage(df, samp:str, constructs:str='all', gene:str=None, cluster:int=None, savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    return 0
    """if constructs == 'all':
        constructs = list(df[df.samp==samp]['construct'].unique())
    trace = [
        go.Scatter(
            x= np.array([i for i in range(len(df.sequence.iloc[0]))]),
            y= np.array([df['min_cov_bases'].iloc[0] for i in range(len(df.sequence.iloc[0]))]) , 
            name='Min coverage bases',
            mode='lines')
    ]
    for construct in constructs:
        mh = Manipulator(df).get_series(df, SubDF.from_locals(locals()))
        x = np.array([i for i in range(len(mh.sequence))])
        y = np.array([int(mh.info_bases[i]) for i in range(len(mh.sequence))])
        trace.append(go.Scatter(
            x=x,
            y=y, 
            name=construct,
            mode='lines'
            ))

    layout = go.Layout(
        title=f"{mh.samp} - {mh.construct} - {mh.cluster}",
        xaxis=dict(title="Bases"),
        yaxis=dict(title="Info bases"),
        plot_bgcolor="white"
        )

    fig = go.Figure(data=trace, layout=layout)
    
    fig = __layout_routine(fig, savefile, auto_open, use_iplot, title)

    return {'fig':fig, 'df':pd.DataFrame({t['name']:{'x':t['x'], 'y':t['y']} for t in trace}).T}"""

def __layout_routine(fig, savefile, auto_open, use_iplot, title):
    if title != None:
        fig['layout']['title'] = title
    if use_iplot:
        iplot(fig)
    if savefile != None:
        plot(fig, filename = savefile, auto_open=auto_open)
    return fig



from dreem.test import files_generator 

number_of_constructs = 1
number_of_reads = [10]*number_of_constructs
mutations = [ [[]]+[[23]]+[[35]]+[[]]*4+[[37]]+[[32]]+[[33,36]] for n in range(number_of_constructs) ] # 0-based
insertions = [ [[]]*3+[[11]]+[[8, 23]]+[[]]*2+[[15]]+[[]]*2 for n in range(number_of_constructs) ] # 0-based
deletions = [ [[]]*5+[[2]]+[[4, 6]]+[[]]+[[8]]+[[]] for n in range(number_of_constructs) ] # 0-based
no_info = [ [[]]*2+[[2]]+[[4, 6]]+[[]]+[[3]]+[[]]*5 for n in range(number_of_constructs) ] # 0-based

#insertions = [ [[]]*10 for n in range(number_of_constructs) ] # 0-based
#no_info = [ [[]]*10 for n in range(number_of_constructs) ] # 0-based
#deletions = [ [[]]*10 for n in range(number_of_constructs) ] # 0-based
#mutations = [ [[]]*10 for n in range(number_of_constructs) ] # 0-based

#number_of_reads = [2]*number_of_constructs
#mutations = [[[10,14], [12]]  for n in range(number_of_constructs) ] # 0-based
#insertions = [ [[], []] for n in range(number_of_constructs) ] # 0-based
#deletions = [ [[4], [ 4,7]]  for n in range(number_of_constructs) ] # 0-based
no_info = [[[]]*number_of_reads[n] for n in range(number_of_constructs) ] # 0-based

length = [50, 150]
sequences = [[files_generator.create_sequence(length[k])]*number_of_reads[k] for k in range(number_of_constructs)]
constructs = ['construct_{}'.format(i) for i in range(number_of_constructs)]
barcode_start = 30
len_barcode = 10
barcodes = files_generator.generate_barcodes(len_barcode, number_of_constructs, 3)
sections_start = [[0, 25],[0, 25, 50, 75]] # 0-based
sections_end = [[25, 50],[25, 50, 75, 99]] # 0-based
sections = [['{}-{}'.format(ss+1, se) for ss,se in zip(sections_start[n], sections_end[n])] for n in range(number_of_constructs)] # 0-based
sample_profile = files_generator.make_sample_profile(constructs, sequences, number_of_reads, mutations, insertions, deletions, no_info, sections=sections, section_start=sections_start, section_end=sections_end, barcodes=barcodes, barcode_start=barcode_start)



import scipy.stats
import numpy as np
from tqdm import tqdm
import pandas as pd

def compute_conf_interval(info_bases, mut_bases, alpha = 0.05):
    assert len(info_bases)==len(mut_bases), "info_bases and mut_bases must be of same length"
    ci = {}
    i, ci['poisson_min'], ci['poisson_max'], ci['poisson_low'], ci['poisson_high'] = 0, np.zeros(len(info_bases)), np.zeros(len(info_bases)), np.zeros(len(info_bases)), np.zeros(len(info_bases))
    for cov, mut in zip(info_bases, mut_bases):
        if cov == 0:
            ci['poisson_min'][i], ci['poisson_max'][i],ci['poisson_low'][i], ci['poisson_high'][i] = np.nan, np.nan, np.nan, np.nan
            continue 
        ci['poisson_min'][i], ci['poisson_max'][i] = 0.5*scipy.stats.chi2.ppf(alpha/2, df=2*mut)/cov, 0.5*scipy.stats.chi2.ppf(1-alpha/2, df=2*(mut+1))/cov
        if mut == 0:
            ci['poisson_min'][i] = 0
        ci['poisson_low'][i], ci['poisson_high'][i] = mut/cov-ci['poisson_min'][i], ci['poisson_max'][i]-mut/cov
        i = i+1
    ci.pop('poisson_min')
    ci.pop('poisson_max')
    ci['poisson_high'] = np.round(ci['poisson_high'], 4).tolist()
    ci['poisson_low'] = np.round(ci['poisson_low'], 4).tolist()
    return ci


def add_poisson_confidence_intervals(cluster, sample, verbose = False):
    ci = {}

    ci[idx] = compute_conf_interval(info_bases=cluster['info_bases'], mut_bases=cluster['mut_bases'])
    
    return df


import numpy as np

import dreem
import click
import os, sys
import pandas as pd
import json
from dreem.util import util as util
sys.path.append(os.path.join(os.path.dirname(__file__)))
from bitvector import BitVector
from clusteringAnalysis import ClusteringAnalysis
from EMclustering import EMclustering
from dreem.util.cli import FASTA, INPUT_DIR, OUT_DIR, MAX_CLUSTERS, MIN_ITER, SIGNAL_THRESH, INFO_THRESH, INCLUDE_G_U, INCLUDE_DEL, MIN_READS, CONVERGENCE_CUTOFF, NUM_RUNS, COORDS, PRIMERS, FILL, N_CPUS, VERBOSE



def run(input_dir:str=INPUT_DIR, out_dir:str=OUT_DIR, max_clusters:int=MAX_CLUSTERS, min_iter:int=MIN_ITER, signal_thresh:float=SIGNAL_THRESH, info_thresh:float=INFO_THRESH, include_g_u:bool=INCLUDE_G_U, include_del:bool=INCLUDE_DEL, min_reads:int=MIN_READS, convergence_cutoff:float=CONVERGENCE_CUTOFF, num_runs:int=NUM_RUNS, n_cpus:int=N_CPUS, verbose:bool=VERBOSE):
    """Run the clustering module.

    Clusters the reads of all given bitvectors and outputs the likelihoods of the clusters as `name`.json in the directory `output_path`, using `temp_path` as a temp directory.
    Each bitvector is a file containing the reads of a construct. Bitvectors from the same sample should be grouped in a folder and the path to the folder should be given as `bv_dir`.
    `name` is the name of the output file, and should be the sample name.

    Parameters from args:
    ---------------------

    input_dir: str
        Path to the bit vector folder or list of paths to the bit vector folders.
    out_dir: str
        Path to the output folder.
    max_clusters: int
        Maximum number of clusters.
    min_iter: int
        Minimum number of iteration per EM exectution.
    signal_thresh: float
        Signal threshold
    info_thresh: float
        Float from 0 to 1, where 1 means that all bases are unvalid and 0 means that all bases are valid (valid:= just 0s and 1s in the bitvector). If info_thresh of a read is above this threshold, it is considered unvalid and isn't used.
    include_g_u: bool
        Include G and U
    include_del: bool
        Include deletions
    min_reads: int
        Minimum number of reads per cluster.
    convergence_cutoff: float
        Convergence cutoff
    num_runs: int
        Number of runs
    n_cpus: int
        Number of cpus
    verbose: bool
        Verbose
        
    Returns
    -------
    
    clustering: dict
        The structure is the following:
        {construct_1}: # name of the bitvector file (from fasta file)
            {section_1}: # name of the clustered section (from library)
                {read_1}:  # read number (from bitvector)
                    K1_1: likelihood that this read belongs to cluster 1 when using 1 cluster
                    K2_1: likelihood that this read belongs to cluster 1 when using 2 clusters
                    K2_2: likelihood that this read belongs to cluster 2 when using 2 clusters
                    K3_1: likelihood that this read belongs to cluster 1 when using 3 clusters
                    K3_2: likelihood that this read belongs to cluster 2 when using 3 clusters
                    K3_3: likelihood that this read belongs to cluster 3 when using 3 clusters   
    
    """

    # Create the output folder
    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(os.path.join(out_dir, 'temp'), exist_ok=True)
    
    # Make sure the input directory is a list
    if type(input_dir) not in (list, tuple):
        input_dir = [input_dir]
    
    # Create the output file
    best_clusters_samples = {}
    
    # Create a clustering args 
    clustering_args = dict(
        max_clusters = max_clusters, 
        min_iter = min_iter,
        signal_thresh = signal_thresh, 
        info_thresh = info_thresh, 
        include_g_u = include_g_u, 
        include_del = include_del, 
        min_reads = min_reads,
        convergence_cutoff = convergence_cutoff,
        num_runs = num_runs,
        n_cpus = n_cpus,
        verbose = verbose
    )
    
    # Get the bitvector files in the input directory and all of its subdirectories
    files_in = []
    for in_dir in input_dir:
        files_in += util.get_files(in_dir, '.orc')
    for i, f_in in enumerate(files_in):
        section = f_in.split('/')[-2]
        print("\n\nSTARTING SAMPLE", i, '|', section)
        bitvector = BitVector(path=f_in)
        bitvector.publish_preprocessing_report(path=os.path.join(out_dir,section+'_preprocessing_report.txt'))
        ca = ClusteringAnalysis(bitvector, max_clusters, num_runs, clustering_args)
        clusters = ca.run()
        reads_best_cluster = {}
        for k in clusters:
            em = EMclustering(bitvector.bv, int(k[1]), bitvector.read_hist, bitvector.base_to_keep, bitvector.sequence, **clustering_args)
            likelihood_reads_best_cluster, _, _ = em.expectation(clusters[k][0]['mu'], clusters[k][0]['pi'])
            reads_best_cluster[k] = bitvector.associate_reads_with_likelihoods(likelihood_reads_best_cluster)
            
        best_clusters_samples[section] = reads_best_cluster

    # Save the results
    with open(os.path.join(out_dir, 'best_cluster_reads.json'), 'w') as f:
        json.dump(best_clusters_samples, f, indent=4)
    
    return 1




from __future__ import annotations
from collections import defaultdict
from functools import cached_property
import itertools
import os
import pathlib
import re
from tqdm import tqdm
from datetime import datetime
from hashlib import file_digest
from multiprocessing import Pool
from typing import Any, List, Optional, Tuple, Dict

import numpy as np
import pandas as pd

from dreem.util.dflt import NUM_PROCESSES
from dreem.util.seq import FastaParser
from dreem.util.path import BasePath, FastaInPath, XamInPath, MutVectorBatchPath, MutVectorBatchSegment, MutVectorReportPath, MutVectorReportSegment, MOD_VEC, TEMP_DIR, OUTPUT_DIR, RegionSegment, RegionPath
from dreem.util.seq import DNA
from dreem.vector.samview import SamViewer
from dreem.vector.vector import SamRecord


DEFAULT_BATCH_SIZE = 33_554_432  # 2^25 bytes ≈ 33.6 Mb


class Region(object):
    __slots__ = ["first", "last", "ref_seq", "ref_name"]

    def __init__(self, ref_name: str, first: int, last: int, ref_seq: DNA):
        """
        Initialize a Region object to represent a region of interest (between a
        first and a last position) in a reference sequence.
        
        ** Arguments **
        ref_name (str) ----> name of the reference
        first (int) ---------> 5'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        last (int) ----------> 3'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        ref_seq (DNA) -------> sequence of the entire reference (not just the
                               region of interest)
        
        ** Returns **
        None
        """
        if first <= 0:
            raise ValueError(f"first ({first}) must be >= 1")
        if last > len(ref_seq):
            raise ValueError(f"last ({last}) must be <= the length of "
                             f"ref_seq ({len(ref_seq)}) {ref_seq}.")
        if -len(ref_seq) <= last < 0:
            # This option allows using non-positive end coordinates to mean
            # distance from the 3' end (similar to Python's indexing), with
            # -1 meaning up to and including the last coordinate, -2 meaning
            # up to an including the second-to-last coordinate, and so on
            # until -len(ref_seq), which means the first coordinate.
            last += len(ref_seq)
        if last < first:
            raise ValueError(f"last ({last}) must be >= first ({first})")
        self.first = first
        self.last = last
        self.ref_seq = ref_seq
        self.ref_name = ref_name
    
    @property
    def spanning(self) -> bool:
        """ Return whether the region spans the entire reference sequence. """
        return self.first == 1 and self.last == len(self.ref_seq)
    
    @property
    def region_seq(self) -> DNA:
        """ Return the sequence of the region of interest. """
        return self.ref_seq[self.first - 1: self.last]

    @property
    def length(self) -> int:
        """ Return the length of the region of interest. """
        return self.last - self.first + 1

    @property
    def positions(self) -> np.ndarray:
        """ Return all positions in the region of interest as an NDArray. """
        return np.arange(self.first, self.last + 1)

    @property
    def ref_coords(self) -> Tuple[str, int, int]:
        """ Return the name of the reference and the first and last positions
        of the region of interest; for equality testing and hashing. """
        return self.ref_name, self.first, self.last
    
    @property
    def columns(self) -> List[str]:
        """ Return a list of the bases and positions in the region of interest,
        each of the form '{base}{position}' (e.g. ['G13', 'C14', 'A15']). """
        # FIXME: add a column for the read names
        return [f"{chr(base)}{pos}" for base, pos
                in zip(self.region_seq, self.positions)]
    
    def __eq__(self, other: object) -> bool:
        """ Two Region objects are equal iff their ref_coords match. """
        if isinstance(other, Region):
            return self.ref_coords == other.ref_coords
        else:
            return NotImplemented        


class RegionFinder(Region):
    # For amplicon-based DMS-MaPseq, primer_gap is the number of positions
    # to leave between the end of the fwd/rev primer in the RT-PCR step
    # and the 5'/3' end of the region. It is used because bases adjacent to
    # primers may show artifacts. Setting it to 0 causes the 5'/3' ends of
    # the region to lie immediately 3'/5' of the fwd/rev primers, respectively.
    # Historically, primer_gap has been set to 2.
    primer_gap = 2

    def __init__(self, ref_name: str, ref_seq: DNA,
                 first: Optional[int] = None, last: Optional[int] = None,
                 fwd: Optional[DNA] = None, rev: Optional[DNA] = None):
        """
        Initialize a RegionFinder object to determine the first and last
        positions in a region. The positions are determined as follows:
        - first
          - If first is given as an argument, this value is used for first.
          - Else if fwd is given, first is based on the location of its 3' end.
          - Else first is set to 1.
        - last
          - If last is given as an argument, last is set to this value.
          - Else if rev is given, last is based on the location of the 5' end
            of its reverse complement.
          - Else last is set to the length of ref_seq.
        
        ** Arguments **
        ref_name (bytes) ----> name of the reference
        ref_seq (DNA) -------> sequence of the entire reference (not just the
                               region of interest)
        first (int) ---------> (optional) 5'-most position of the reference
                               that lies in the region of interest
        last (int) ----------> (optional) 3'-most position of the reference
                               that lies in the region of interest
        fwd (DNA) -----------> (optional) sequence of the forward primer used
                               to amplify the region (for amplicons only)
        rev (DNA) -----------> (optional) sequence of the reverse primer used
                               to amplify the region (for amplicons only)
        
        ** Returns **
        None
        """
        if first is None:
            # If first is to be determined from the fwd primer sequence,
            # the primer is aligned to the reference, and first is set to
            # the position (primer_gap + 1) downstream of its 3' coordinate.
            first = (1 if fwd is None else self.locate_primer(
                ref_seq, fwd)[1] + self.primer_offset)
        if last is None:
            # If last is to be determined from the rev primer sequence,
            # the reverse complement of the primer is aligned to the reference,
            # and last is set to the position (primer_gap + 1) upstream of its
            # 5' coordinate.
            last = (len(ref_seq) if rev is None else self.locate_primer(
                ref_seq, rev.rc)[0] - self.primer_offset)
        super().__init__(ref_name, first, last, ref_seq)

    @property
    def primer_offset(self):
        return self.primer_gap + 1

    @staticmethod
    def locate_primer(ref_seq: DNA, primer: DNA) -> tuple[int, int]:
        """
        Return the 5' and 3' positions (1-indexed) of a primer within a
        reference sequence. The primer must occur exactly once in the
        reference, otherwise an error is raised.

        ** Arguments **
        ref_seq (DNA) -> sequence of the reference
        primer (DNA) --> sequence of the forward primer, or of the reverse
                         complement of the reverse primer
        
        ** Returns **
        pos5 (int) ----> 5'-most position of the primer aligned to the reference
        pos3 (int) ----> 3'-most position of the primer aligned to the reference
        """
        matches = list(re.finditer(primer, ref_seq))
        if not matches:
            raise ValueError(f"Primer '{primer}' is not in ref '{ref_seq}'")
        if len(matches) > 1:
            raise ValueError(f"Primer '{primer}' occurs {len(matches)} times "
                             f"in ref '{ref_seq}'")
        # Add 1 to convert from 0-indexed (re.finditer) to 1-indexed (DREEM).
        pos5 = matches[0].start() + 1
        # No change is needed to convert from exclusive 0-indexed (re.finditer)
        # to inclusive 1-indexed (DREEM).
        pos3 = matches[0].end()
        return pos5, pos3
    

class MutationalProfile(Region):
    __slots__ = ["sample_name"]

    def __init__(self, sample_name: str, ref_name: str,
                 first: int, last: int, ref_seq: DNA):
        """
        Initialize a MutationalProfile object that represents all of the reads
        from a particular sample that overlap a region of interest.
        
        ** Arguments **
        sample_name (str) ---> name of the sample
        ref_name (str) ----> name of the reference
        first (int) ---------> 5'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        last (int) ----------> 3'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        ref_seq (DNA) -------> sequence of the entire reference (not just the
                               region of interest)
        
        ** Returns **
        None
        """
        super().__init__(ref_name, first, last, ref_seq)
        self.sample_name = sample_name
    
    def __str__(self) -> str:
        return (f"Mutational Profile of sample '{self.sample_name}' reference"
                f" '{self.ref_name}' region {self.first}-{self.last}")


class VectorIO(MutationalProfile):
    __slots__ = ["base_path", "num_batches", "num_vectors", "checksums"]

    digest_algo = "md5"

    def __init__(self, base_path: BasePath, sample_name: str, ref_name: str,
                 first: int, last: int, ref_seq: DNA):
        """
        Initialize a VectorIO object to read and write mutation vectors.
        
        ** Arguments **
        base_path (str) -------> path to directory in which the output files
                               will be written ({base_path}/{sample_name}/
                               {ref_name}/{first}-{last})
        sample_name (str) ---> name of the sample
        ref_name (bytes) ----> name of the reference
        first (int) ---------> 5'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        last (int) ----------> 3'-most position of the reference that lies in
                               the region of interest (1-indexed, inclusive)
        ref_seq (DNA) -------> sequence of the entire reference (not just the
                               region of interest)

        ** Returns **
        None
        """
        super().__init__(sample_name, ref_name, first, last, ref_seq)
        self.base_path = base_path
        self.num_batches = 0
        self.num_vectors = 0
        self.checksums = list()
    
    @property
    def args(self):
        return (self.base_path.path,
                OUTPUT_DIR,
                MOD_VEC,
                self.sample_name,
                self.ref_name,
                RegionSegment.format(self.first, self.last))

    @property
    def report_path(self):
        mv_report = MutVectorReportSegment.format(self.first, self.last)
        args = self.args[:-1] + (mv_report,)
        return MutVectorReportPath(*args)

    @property
    def batch_dir(self):
        return RegionPath(*self.args)

    def get_mv_batch_path(self, batch_num: int):
        mv_batch = MutVectorBatchSegment.format(batch_num)
        args = self.args + (mv_batch,)
        return MutVectorBatchPath(*args)
    
    @property
    def batch_nums(self):
        """ List all of the batch numbers. """
        return range(self.num_batches)

    @property
    def mv_batch_paths(self):
        return list(map(self.get_mv_batch_path, self.batch_nums))

    @classmethod
    def digest_file(cls, path: str) -> str:
        """
        Compute the MD5 checksum of a file.
        
        ** Arguments **
        path (str) ---> path of the file
        
        ** Returns **
        digest (str) <- MD5 checksum of the file
        """
        with open(path, "rb") as f:
            digest = file_digest(f, cls.digest_algo).hexdigest()
        return digest


class Report(VectorIO):
    __slots__ = ["began", "ended"]

    # fields is a dict that maps the name of each field to its data type
    # and defines the order of the fields in the report file.
    fields = {"Sample Name": str, "Ref Name": str, "First": int, "Last": int,
              "Ref Seq": DNA, "Num Batches": int, "Num Vectors": int,
              "Checksums": list, "Began": datetime, "Ended": datetime,
              "Duration": float, "Speed": float}
    
    # units is a dict that defines the units of several fields that have them.
    units = {"Speed": "vec/s", "Duration": "s",
             "Checksums": VectorIO.digest_algo}

    # format of dates and times in the report file
    datetime_fmt = "on %Y-%m-%d at %H:%M:%S.%f"

    def __init__(self, base_path: str, sample_name: str, ref_name: str,
                 first: int, last: int, ref_seq: DNA, num_batches: int,
                 num_vectors: int, checksums: List[str],
                 began: datetime, ended: datetime):
        """
        Initialize a Report object to
        - record information about a mutational profile and its mutation vectors
          in a human-readable plain-text format
        - facilitate validating and loading ORC files of the mutation vectors
          during subsequent steps of DREEM
        
        ** Arguments **
        base_path (str) ---------> path to directory in which the output files
                                 will be written ({base_path}/{sample_name}/
                                 {ref_name}/{first}-{last})
        sample_name (str) -----> name of the sample
        ref_name (str) ------> name of the reference
        first (int) -----------> 5'-most position of the reference that lies in
                                 the region of interest (1-indexed, inclusive)
        last (int) ------------> 3'-most position of the reference that lies in
                                 the region of interest (1-indexed, inclusive)
        ref_seq (DNA) ---------> sequence of the entire reference (not just the
                                 region of interest)
        num_batches (int) -----> number of batches in the mutational profile
        num_vectors (int) -----> number of vectors in the mutational profile
        checksums (list[str]) -> list of checksums for mutation vector files

        ** Returns **
        None
        """
        super().__init__(base_path, sample_name, ref_name, first, last, ref_seq)
        self.num_batches = num_batches
        self.num_vectors = num_vectors
        self.checksums = checksums
        assert ended >= began
        self.began = began
        self.ended = ended
    
    @property
    def duration(self) -> float:
        """ Return duration of the computation (in seconds) """
        dt = self.ended - self.began
        return dt.seconds + dt.microseconds / 1E6
    
    @property
    def speed(self) -> float:
        """ Return speed of the computation (in vectors per second) """
        try:
            return self.num_vectors / self.duration
        except ZeroDivisionError:
            return float("inf" if self.num_vectors else "nan")

    @classmethod
    def append_unit(cls, field: str) -> str:
        """ Append the unit to the name of a field. """
        return f"{field} ({unit})" if (unit := cls.units.get(field)) else field

    @classmethod
    def remove_unit(cls, field: str) -> str:
        """ Remove the unit from a field, leaving only its name.
        NOTE: This method only works assuming that every field that has a unit
        (namely "Speed" and "Duration") contains no whitespace. This method
        will need rewriting if this assumption changes in the future. """
        return word if cls.units.get(word := field.split(" ")[0]) else field

    @classmethod
    def attr_to_field(cls, attr: str):
        """ Convert the name of an attribute of a Report object (in snake_case)
        to the name of a field in the report file (in Capital Case). """
        return cls.append_unit(" ".join(map(str.capitalize, attr.split("_"))))

    @classmethod
    def field_to_attr(cls, field: str):
        """ Convert the name of a field in the report file (in Capital Case)
        to the name of an attribute of a Report object (in snake_case). """
        return cls.remove_unit(field).replace(' ', '_').lower()

    @classmethod
    def format_val(cls, val: Any) -> str:
        """ Return a string representation of the value of a field. """
        dtype = type(val)
        if dtype is str or dtype is int or dtype is DNA:
            return val
        if dtype is float:
            return round(val, 2)
        if dtype is datetime:
            return val.strftime(cls.datetime_fmt)
        if dtype is list:
            return ", ".join(val)
        raise ValueError(dtype)

    @classmethod
    def parse_valstr(cls, field: str, valstr: str) -> Any:
        """ Return the value of a field from a string representation. """
        dtype = cls.fields[cls.remove_unit(field)]
        if dtype is str or dtype is int or dtype is float:
            return dtype(valstr)
        if dtype is DNA:
            return DNA(valstr.encode())
        if dtype is datetime:
            return datetime.strptime(valstr, cls.datetime_fmt)
        if dtype is list:
            return valstr.split(", ")
        raise ValueError(dtype)

    def save(self):
        """ Save the information in a Report to a file. """
        # Determine the maximum number of characters in a label.
        width = max(map(len, map(self.append_unit, self.fields.keys())))
        # Create a format string that pads every label to that length.
        pattern = "{field: <" + str(width) + "}\t{val}\n"
        lines: List[str] = list()
        # Iterate through all the fields to include in the report.
        for field in self.fields.keys():
            # Get the name of the Report attribute corresponding to the field.
            attr = self.field_to_attr(field)
            # Get a string representation of the value of the field.
            valstr = self.format_val(self.__getattribute__(attr))
            # Format the line of the report with the field's name and value.
            line = pattern.format(field=self.append_unit(field), val=valstr)
            lines.append(line)
        # Write the lines to the report file.
        with open(self.report_path.path, "w") as f:
            f.write("".join(lines))
    
    @classmethod
    def load(cls, report_file: str) -> Report:
        """ Return a Report from a saved report file. """
        # FIXME: make this function to find base_path more elegant
        region_dir = os.path.dirname(os.path.abspath(report_file))
        ref_dir = os.path.dirname(region_dir)
        sample_dir = os.path.dirname(ref_dir)
        module_dir = os.path.dirname(sample_dir)
        base_path = os.path.dirname(module_dir)
        # Initialize the dict of attribute values with base_path.
        vals: Dict[str, Any] = {"base_path": base_path}
        with open(report_file) as f:
            for line in f:
                # Read the field and the string representation of its value.
                field, valstr = map(str.rstrip, line.split("\t"))
                # Get the name of the attribute corresponding to the field.
                attr = cls.field_to_attr(field)
                # Parse the string representation and set the attribute value.
                vals[attr] = cls.parse_valstr(field, valstr)
        # Return a new Report object from the attributes and their values.
        return cls(**vals)


class VectorWriter(VectorIO):
    __slots__ = ["bam_path", "parallel_reads", "region_seqb"]

    """
    Computes mutation vectors for all reads from one sample mapping to one
    region of one reference sequence.
    """
    def __init__(self, base_path: BasePath, bam_path: XamInPath, ref_name: str,
                 first: int, last: int, ref_seq: DNA, parallel_reads: bool):
        sample = bam_path.sample.name
        super().__init__(base_path, sample, ref_name, first, last, ref_seq)
        self.bam_path = bam_path
        self.parallel_reads = parallel_reads
        self.region_seqb = bytes(self.region_seq)

    def _write_batch(self, read_names: Tuple[str], muts: Tuple[bytearray],
                     batch_num: int) -> Tuple[pathlib.Path, int]:
        """
        Write a batch of mutation vectors to an ORC file.

        ** Arguments **
        muts (NDArray) --> batch of mutation vectors in which each row is a
                           mutation vector and each column is a position in the
                           region of interest
        batch_num (int) -> non-negative integer label for the batch; every
                           mutational profile containing n batches includes all
                           batch numbers i in the range 0 <= i < n

        ** Returns **
        mv_file (str) <--- file path where the mutation vectors were written
        """
        assert batch_num >= 0
        # Convert the concatenated mutation vectors to a 2D NumPy array.
        muts_array = np.frombuffer(b"".join(muts), dtype=np.byte)
        n_records = len(muts_array) // self.length
        muts_array.resize((n_records, self.length))
        # Data must be converted to pd.DataFrame for PyArrow to write.
        # Explicitly set copy=False to copying the mutation vectors.
        df = pd.DataFrame(data=muts_array, index=read_names,
                          columns=self.columns, copy=False)
        mv_file = self.get_mv_batch_path(batch_num).path
        df.to_orc(mv_file, engine="pyarrow")
        return mv_file, n_records

    def _write_report(self, t_start: datetime, t_end: datetime):
        Report(self.base_path, self.sample_name, self.ref_name, self.first,
               self.last, self.ref_seq, self.num_batches, self.num_vectors,
               self.checksums, t_start, t_end).save()

    def _vectorize_record(self, rec: SamRecord):
        """
        Compute the mutation vector of one record from a SAM file.

        ** Arguments **
        rec (SamRecord) --> SAM record for which to compute a mutation vector

        ** Returns **
        muts (bytearray) <- mutation vector
        """
        assert rec.ref_name == self.ref_name
        muts = rec.vectorize(self.region_seqb, self.first, self.last)
        assert any(muts)
        return rec.read_name, muts

    def _vectorize_batch(self, sam_viewer: SamViewer, batch_num: int,
                         start: int, stop: int):
        """
        Generate a batch of mutation vectors and write them to an ORC file.

        ** Arguments **
        sam_viewer (SamViewer) -> viewer to the SAM file for which to generate
                                  a batch of mutation vectors
        batch_num (int) --------> non-negative integer label for the batch
        start (int) ------------> seek to this position in the SAM file, then
                                  start generating vectors; position must be
                                  the beginning of a line, otherwise integrity
                                  checks will fail
        stop (int) -------------> stop generating vectors upon reaching this
                                  position in the SAM file; position must be
                                  the beginning of a line, otherwise integrity
                                  checks will fail
        
        ** Returns **
        n_records (int) <-------- number of records read from the SAM file
                                  between positions start and stop
        checksum (str) <--------- MD5 checksum of the ORC file of vectors
        """
        if stop > start:
            with sam_viewer as sv:
                # Use the SAM viewer to generate the mutation vectors.
                # Collect them as a single, 1-dimensional bytes object.
                read_names, muts = zip(*map(self._vectorize_record,
                                            sv.get_records(start, stop)))
        else:
            read_names, muts = (), ()
            raise Warning(f"{self} contained no reads.")
        # Write the mutation vectors to a file and compute its checksum.
        mv_file, n_records = self._write_batch(read_names, muts, batch_num)
        checksum = self.digest_file(mv_file)
        return n_records, checksum

    def _vectorize_sam(self):
        with SamViewer(self.base_path, self.bam_path, self.ref_name,
                       self.first, self.last, self.spanning) as sv:
            batch_size = max(1, DEFAULT_BATCH_SIZE // self.length)
            indexes = list(sv.get_batch_indexes(batch_size))
            starts = indexes[:-1]
            stops = indexes[1:]
            self.num_batches = len(starts)
            assert self.num_batches == len(stops)
            svs = [SamViewer(self.base_path, sv.sam_path, self.ref_name,
                             self.first, self.last, self.spanning, owner=False)
                   for _ in self.batch_nums]
            args = list(zip(svs, self.batch_nums, starts, stops))
            if self.parallel_reads:
                n_procs = max(1, min(NUM_PROCESSES, self.num_batches))
                with Pool(n_procs, maxtasksperchild=1) as pool:
                    results = pool.starmap(self._vectorize_batch, args,
                                           chunksize=1)
            else:
                results = list(itertools.starmap(self._vectorize_batch, args))
            assert len(results) == self.num_batches
            assert self.num_vectors == 0
            assert len(self.checksums) == 0
            for num_vectors, checksum in results:
                self.num_vectors += num_vectors
                self.checksums.append(checksum)
    
    def vectorize(self):
        if not (all(f.path.is_file() for f in self.mv_batch_paths)
                and self.report_path.path.is_file()):
            self.batch_dir.path.mkdir(parents=True, exist_ok=True)
            print(f"{self}: computing vectors")
            t_start = datetime.now()
            self._vectorize_sam()
            t_end = datetime.now()
            print(f"{self}: writing report")
            self._write_report(t_start, t_end)
            print(f"{self}: finished")


class VectorWriterSpawner(object):
    def __init__(self,
                 base_dir: str,
                 fasta: str,
                 bam_files: List[str],
                 coords: List[Tuple[str, int, int]],
                 primers: List[Tuple[str, DNA, DNA]],
                 fill: bool,
                 parallel: str):
        self.base_path = BasePath.parse(base_dir)
        self.bam_paths = list(map(XamInPath.parse, bam_files))
        self.ref_path = FastaInPath.parse(fasta)
        self.coords = coords
        self.primers = primers
        self.fill = fill
        if parallel == "auto":
            parallel = ("reads" if self.num_samples == self.num_regions == 1
                        else "profiles")
        if parallel == "profiles":
            self.parallel_profiles = True
            self.parallel_reads = False
        elif parallel == "reads":
            self.parallel_profiles = False
            self.parallel_reads = True
        elif parallel == "off":
            self.parallel_profiles = False
            self.parallel_reads = False
        else:
            raise ValueError(f"Invalid value for parallel: '{parallel}'")
    
    @property
    def bams_per_sample(self):
        samples: Dict[str, List[XamInPath]] = defaultdict(list)
        for bam in self.bam_paths:
            samples[bam.sample].append(bam)
        return samples
    
    @property
    def samples(self):
        return set(self.bams_per_sample.keys())
    
    @property
    def num_samples(self):
        return len(self.samples)
    
    @property
    def num_regions(self):
        return len(self.coords) + len(self.primers)
    
    @cached_property
    def ref_seqs(self):
        seqs = dict(FastaParser(self.ref_path.path).parse())
        if not seqs:
            raise ValueError(f"'{self.ref_path}' contained no sequences")
        return seqs
    
    @cached_property
    def regions(self):
        regions: Dict[bytes, List[RegionFinder]] = defaultdict(list)

        def add_region(region: RegionFinder):
            if any(region == other for other in regions[region.ref_name]):
                raise ValueError(f"Duplicate region: {region.ref_coords}")
            regions[region.ref_name].append(region)
        
        def ref_to_seq(ref_seqs: Dict[bytes, DNA], ref: bytes):
            try:
                return ref_seqs[ref]
            except KeyError:
                raise ValueError(f"No reference named '{ref.decode()}'")

        for ref, first, last in self.coords:
            add_region(RegionFinder(ref, ref_to_seq(self.ref_seqs, ref),
                                    first=first, last=last))
        for ref, fwd, rev in self.primers:
            add_region(RegionFinder(ref, ref_to_seq(self.ref_seqs, ref),
                                    fwd=fwd, rev=rev))
        if self.fill:
            for ref, seq in self.ref_seqs.items():
                if ref not in regions:
                    add_region(RegionFinder(ref, seq))
        return regions

    @property
    def writers(self):
        for bam in self.bam_paths:
            print(bam, bam.xam._segment)
            ref_name = bam.xam.name
            for region in self.regions[ref_name]:
                assert region.ref_name == ref_name
                yield VectorWriter(self.base_path, bam, ref_name,
                                   region.first, region.last, region.ref_seq,
                                   self.parallel_reads)

    def profile(self, processes: int = 0):
        writers = list(self.writers)
        if not writers:
            raise ValueError("No samples and/or regions were given.")
        if self.parallel_profiles:
            with Pool(processes if processes
                      else min(NUM_PROCESSES, len(writers)),
                      maxtasksperchild=1) as pool:
                pool.map(VectorWriter.vectorize, writers,
                         chunksize=1)
        else:
            for writer in writers:
                writer.vectorize()


'''
class VectorReader(VectorIO):
    @property
    def shape(self):
        return self.num_vectors, self.length

    def run_checksums(self):
        if len(mv_files := self.mv_files) != len(self.checksums):
            raise ValueError(f"Got {len(mv_files)} files but "
                             f"{len(self.checksums)} checksums")
        for mv_file, checksum in zip(mv_files, self.checksums):
            digest = self.digest_file(mv_file)
            if digest != checksum:
                raise ValueError(f"Hex digest of {mv_file} ({digest}) "
                                 f"did not match checksum ({checksum})")

    @property
    def vectors(self):
        self.run_checksums()
        # FIXME: I suspect there are more efficient ways to load these files
        mvs = np.full(self.shape, fill_value=BLANK, dtype=bytes)
        row = 0
        for mv_file in self.mv_files:
            data = orc.read_table(mv_file).to_pandas().values
            n_vectors, n_cols = data.shape
            if n_cols != self.length:
                raise ValueError(f"Expected DataFrame with {self.length}"
                                 f" columns but got {n_cols} columns.")
            mvs[row: (row := row + n_vectors)] = data
        if row != self.num_vectors:
            raise ValueError(f"Expected DataFrame with {self.num_vectors}"
                             f" rows but got {row} rows.")
        vectors = pd.DataFrame(data=mvs, columns=self.positions)
        return VectorSet(self.sample_name, self.ref_name, self.first,
                         self.last, self.ref_seq, vectors)

    @classmethod
    def load(cls, report_file):
        rep = Report.load(report_file)
        return cls(rep.base_path, rep.sample_name, rep.ref_name, rep.first,
                   rep.last, rep.ref_seq, rep.num_batches, rep.num_vectors,
                   rep.checksums)


class VectorSet(MutationalProfile):
    __slots__ = ["_vectors", "_cover_count", "_mm_count", "_del_count"]

    color_dict = {"A": "red", "C": "blue", "G": "orange", "T": "green"}

    def __init__(self, sample_name: str, ref_name: str, first: int, last: int,
                 ref_seq: DNA, vectors: pd.DataFrame):
        super().__init__(sample_name, ref_name, first, last, ref_seq)
        if (vectors.shape[1] != self.length
                or (vectors.columns != self.positions).any()):
            raise ValueError("Columns of vectors do not match positions.")
        self._vectors = vectors
        self._cover_count = None
        self._mm_count = None
        self._del_count = None

    @staticmethod
    def series(f):
        def wrapper(self):
            return pd.Series(f(self), index=self.positions)
        return wrapper

    @property
    def vectors(self):
        return self._vectors

    @property
    def num_vectors(self):
        return self.vectors.shape[0]

    @property
    def mismatch_frac(self):
        return self.mismatch_count / self.coverage_count

    @property
    @series
    def deletion_count(self):
        if self._del_count is None:
            self._del_count = (self.vectors.values == DELET).sum(axis=0)
        return self._del_count

    @property
    def deletion_frac(self):
        return self.deletion_count / self.coverage_count

    @property
    def mutation_count(self):
        return self.mismatch_count + self.deletion_count

    @property
    def mutation_frac(self):
        return self.mutation_count / self.coverage_count

    @property
    def colors(self):
        return [self.color_dict[chr(base)] for base in self.region_seq]

    def plot_muts(self):
        fig, ax = plt.subplots()
        data = self.mutation_frac
        ax.bar(data.index, data, color=self.colors)
        plt.show()
'''



import logging
import os
from multiprocessing import Pool
from typing import Optional

from dreem.util.cli import DEFAULT_INTERLEAVED_INPUT, DEFAULT_INTERLEAVE_OUTPUT, DEFAULT_TRIM, DEFAULT_NEXTSEQ_TRIM
from dreem.util.dflt import NUM_PROCESSES
from dreem.util.seq import FastaParser, FastaWriter
from dreem.util.reads import FastqAligner, FastqTrimmer, get_demultiplexed_fastq_pairs, FastqUnit, BamAlignSorter, SamRemoveEqualMappers, BamSplitter
from dreem.util.stargs import starstarmap
from dreem.util.path import TEMP_DIR, try_remove, BasePath, FastaSampleOutPath, MOD_ALN, FastaSegment, FastaInPath
from dreem.util.seq import DNA


def _align(base_path: BasePath,
           ref_file: FastaInPath | FastaSampleOutPath,
           fq_unit: FastqUnit,
           trim: bool=DEFAULT_TRIM,
           nextseq_trim: bool=DEFAULT_NEXTSEQ_TRIM):
    # Trim the FASTQ file(s).
    if trim:
        trimmer = FastqTrimmer(base_path, fq_unit)
        fq_unit = trimmer.run(nextseq_trim=nextseq_trim)
    # Align the FASTQ to the reference.
    aligner = FastqAligner(base_path, fq_unit, ref_file)
    xam_path = aligner.run()
    trimmer.clean()
    # Remove equally mapping reads.
    remover = SamRemoveEqualMappers(base_path, xam_path)
    xam_path = remover.run()
    aligner.clean()
    # Sort the SAM file and output a BAM file.
    sorter = BamAlignSorter(base_path, xam_path)
    xam_path = sorter.run()
    remover.clean()
    # Split the BAM file into one file for each reference.
    splitter = BamSplitter(base_path, xam_path, ref_file)
    bams = splitter.run()
    sorter.clean()
    return bams


def all_refs(base_dir: str, ref_file: str,
             fastqu: str, fastqi: str, fastq1: str, fastq2: str, **kwargs):
    base_path = BasePath.parse(base_dir)
    ref_path = FastaInPath.parse(ref_file)
    fq_unit = FastqUnit(fastqu=fastqu, fastqi=fastqi, fastq1=fastq1, fastq2=fastq2)
    return _align(base_path, ref_path, fq_unit, **kwargs)


def _get_fq_inputs(fastqu_dir: str, fastqi_dir: str, fastq12_dir: str):
    def assert_inputs_exist(exist: bool):
        if bool(fq_inputs) != exist:
            error = "no" if exist else "more than one"
            raise ValueError(f"Received {error} argument for FASTQ directory.")
    fq_inputs = dict()
    if fastqu_dir:
        assert_inputs_exist(False)
        # FIXME
    if fastqi_dir:
        assert_inputs_exist(False)
        # FIXME
    if fastq12_dir:
        assert_inputs_exist(False)
        fq_inputs = get_demultiplexed_fastq_pairs(fastq12_dir)
    assert_inputs_exist(True)
    return fq_inputs


def _one_ref(base_dir: str, ref: str, seq: DNA, fq_unit: FastqUnit, **kwargs):
    # Write a temporary FASTA file for the reference.
    base_path = BasePath.parse(base_dir)
    fasta_name = FastaSegment.format(ref)
    fasta = FastaSampleOutPath(base_dir, TEMP_DIR, MOD_ALN, fq_unit.sample,
                               fasta_name)
    fasta.path.parent.mkdir(parents=True, exist_ok=True)
    try:
        FastaWriter(fasta.path, {ref: seq}).write()
        bams = _align(base_path, fasta, fq_unit, **kwargs)
        assert len(bams) == 1
        return bams[0]
    finally:
        try_remove(fasta.path)


def each_ref(base_dir: str, refs_file: str,
             fastqu_dir: str, fastqi_dir: str, fastq12_dir: str, **kwargs):
    base_path = BasePath.parse(base_dir)
    fq_inputs = _get_fq_inputs(fastqu_dir, fastqi_dir, fastq12_dir)
    align_args = list()
    align_kwargs = list()
    for ref, seq in FastaParser(refs_file).parse():
        try:
            fq_unit = fq_inputs[ref]
        except KeyError:
            logging.warning(f"No FASTQ files for reference '{ref}'")
        else:
            align_args.append((base_path, ref, seq, fq_unit))
            align_kwargs.append(kwargs)
            
    assert len(align_args) == len(align_kwargs)
    if align_args:
        n_procs = min(len(align_args), NUM_PROCESSES)
        with Pool(n_procs) as pool:
            bams = starstarmap(pool.starmap, _one_ref, align_args, align_kwargs)
    return bams



import click
from pipeline import run
from click_option_group import optgroup
from dreem.util.cli import *

@click.command()

@optgroup.group('I/O')
@opto_out_dir
@fasta
@fastq1
@fastq2
@input_dir
@samples
@sample
@clustering_file
@opti_library    

@optgroup.group('Selection')
@coords
@primers
@fill
@parallel

@optgroup.group('Demultiplexing')
@demultiplexing
@barcode_start
@barcode_length
@max_barcode_mismatches

@optgroup.group('Clustering')
@clustering
@max_clusters
@min_iter
@signal_thresh
@info_thresh
@include_g_u
@include_del
@min_reads
@convergence_cutoff
@num_runs
@n_cpus

@optgroup.group('Aggregation')
@rnastructure_path
@rnastructure_temperature
@rnastructure_fold_args
@rnastructure_dms
@rnastructure_dms_min_unpaired_value
@rnastructure_dms_max_paired_value
@rnastructure_partition
@rnastructure_probability
@poisson

@optgroup.group('Misc')
@verbose

def cli(**args):
    run(**args)

if __name__ == '__main__':
    run()




"""Init file """
__version__= '0.0.5'

import sys, os

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

import dreem.demultiplex.main as demultiplexing
import dreem.align.main as alignment
import dreem.vector.main as vectoring
import dreem.cluster.main as clustering
import dreem.aggregate.main as aggregation
import dreem.draw.main as drawer

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)




import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import dreem
import dreem.util.util as util
import pandas as pd
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import pytest
import numpy as np

# Function creating lists of mutations, insertions or deletions
def get_change(n_change, bc_pos, seq_length):

    n_change_before_bc = np.round( n_change*bc_pos[0] / (seq_length - (bc_pos[1]-bc_pos[0])) ).astype(np.int64)

    change_before = np.round(np.linspace(2, bc_pos[0]-2, n_change_before_bc)).astype(np.int64)
    change_after = np.round(np.linspace(bc_pos[1]+1, seq_length-2, n_change-n_change_before_bc)).astype(np.int64)

    assert (np.diff(change_before) > 1).all()
    assert (np.diff(change_after) > 1).all()

    return [list(change_before)+list(change_after)]


# Code for test set 1
sample_name_1 = 'test_set_1'
module = 'alignment'
number_of_constructs = 1
number_of_reads = [23]
mutations = [ [[]]*20+[[2, 26, 42]]+[[5, 8, 25, 35, 47]]+[[2, 8, 22, 25, 28, 31, 35, 41, 45, 48]] ]
length = 50
reads = [[files_generator.create_sequence(length)]*number_of_reads[k] for k in range(number_of_constructs)]
insertions = [[[]]*number_of_reads[k] for k in range(number_of_constructs)]
deletions = [[[]]*number_of_reads[k] for k in range(number_of_constructs)]
no_info = [[[]]*number_of_reads[k] for k in range(number_of_constructs)]

constructs = ['construct_{}'.format(i) for i in range(number_of_constructs)]
barcode_start = 10
barcodes = files_generator.generate_barcodes(10, number_of_constructs, 3)
sections_start = [[0]]*number_of_constructs
sections_end = [[5]]*number_of_constructs
sections = [['{}-{}'.format(ss+1, se) for ss,se in zip(sections_start[n], sections_end[n])] for n in range(number_of_constructs)]

sample_profile_1 = files_generator.make_sample_profile(constructs, reads, number_of_reads, mutations, insertions, deletions, no_info, sections=sections, section_start=sections_start, section_end=sections_end, barcodes=barcodes, barcode_start=barcode_start)



# Code for test set 2
sample_name_2 = 'test_set_2'
seq_ls = [50, 150, 600]*3+[50]
number_of_constructs = 10
n_reads = [23, 10, 10, 10, 23, 23, 23, 100, 1, 12]
barcode_start = 10
barcode_len = 8
bc_pos = (barcode_start, barcode_start+barcode_len)

# Generic mutation lists
default_mut = lambda i : [[]]*20 + get_change(3, bc_pos, seq_ls[i]) + get_change(5, bc_pos, seq_ls[i]) + get_change(10, bc_pos, seq_ls[i])
default_del_insert_small = lambda i : [[]]*4 + get_change(1, bc_pos, seq_ls[i])*3 + get_change(2, bc_pos, seq_ls[i])*2 + get_change(3, bc_pos, seq_ls[i])
default_del_insert_large = lambda i : [[]]*20 + get_change(1, bc_pos, seq_ls[i]) + get_change(2, bc_pos, seq_ls[i]) + get_change(3, bc_pos, seq_ls[i])

mutations = [
    default_mut(0),
    [[]]*n_reads[1] ,
    [[]]*n_reads[2] ,
    [[]]*n_reads[3] ,
    default_mut(4),
    default_mut(5),
    default_mut(6),
    [[]]*95 + get_change(5, bc_pos, seq_ls[7])*5,
    [[]]*n_reads[8] ,
    [[]]*n_reads[9] ,
]

deletions = [
    [[]]*n_reads[0],
    default_del_insert_small(1),
    [[]]*n_reads[2] ,
    default_del_insert_small(3),
    default_del_insert_large(4),
    [[]]*n_reads[5],
    default_del_insert_large(6),
    [[]]*n_reads[7],
    [[]]*n_reads[8] ,
    [[]]*n_reads[9] ,
]

insertions = [
    [[]]*n_reads[0],
    [[]]*n_reads[1],
    default_del_insert_small(2),
    default_del_insert_small(3),
    [[]]*n_reads[4] ,
    default_del_insert_large(5),
    default_del_insert_large(6),
    [[]]*n_reads[7],
    [[]]*n_reads[8] ,
    [[]]*n_reads[9] ,
]

no_info = [
    [[]]*n_reads[0],
    [[]]*n_reads[1],
    default_del_insert_small(2),
    default_del_insert_small(3),
    [[]]*n_reads[4] ,
    default_del_insert_large(5),
    [[]]*n_reads[6] ,
    [[]]*n_reads[7],
    [[]]*n_reads[8] ,
    [[]]*n_reads[9] ,
]

for i in range(number_of_constructs):
    for j in range(n_reads[i]):
        for m in mutations[i][j]:
            if m in deletions[i][j]:
                deletions[i][j].remove(m)
            if m in insertions[i][j]:
                insertions[i][j].remove(m)
            if m in no_info[i][j]:
                no_info[i][j].remove(m)
        for d in deletions[i][j]:
            if d in insertions[i][j]:
                insertions[i][j].remove(d)
            if d in no_info[i][j]:
                no_info[i][j].remove(d)

reads = [[files_generator.create_sequence(seq_ls[k])]*n_reads[k] for k in range(number_of_constructs)]
constructs = ['construct_{}'.format(i) for i in range(number_of_constructs)]
barcodes = files_generator.generate_barcodes(barcode_len, number_of_constructs, 3)
sections_start = [[0]]*number_of_constructs
sections_end = [[5]]*number_of_constructs
sections = [['{}-{}'.format(ss+1, se) for ss,se in zip(sections_start[n], sections_end[n])] for n in range(number_of_constructs)]

sample_profile_2 = files_generator.make_sample_profile(constructs, reads, n_reads, mutations, insertions, deletions, no_info, sections=sections, section_start=sections_start, section_end=sections_end, barcodes=barcodes, barcode_start=barcode_start)


sample_profiles = [sample_profile_1, sample_profile_2]
sample_names = [sample_name_1, sample_name_2]

module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

inputs = ['fastq','fasta']
outputs = ['bam']

# ### Create test files for `test set 1`
def test_make_files():
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'output', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'input', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'expected_output', module)))
    if not os.path.exists(os.path.join(test_files_dir, 'input', module)):
        os.makedirs(os.path.join(test_files_dir, 'input', module))
    
    for sample_profile, sample_name in zip(sample_profiles, sample_names):
        files_generator.generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name)
        files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
        files_generator.assert_files_exist(sample_profile, module, outputs, prediction_dir, sample_name)

#@pytest.mark.skip(reason="Dependencies not implemented yet")
def test_run():
    for sample in os.listdir(module_input):
        dreem.alignment.run(
            out_dir = os.path.dirname(output_dir),
            fastq = '{}/{}_R1.fastq'.format(os.path.join(module_input,sample),sample),
            fastq2 = '{}/{}_R2.fastq'.format(os.path.join(module_input,sample),sample),
            fasta = '{}/reference.fasta'.format(os.path.join(module_input,sample)),
            demultiplexing = False,
            nextseq_trim=False
        )
        
#def test_copy_prediction_as_results():
#    files_generator.copy_prediction_as_results(module_expected, os.path.join(module_output,'output'))

#@pytest.mark.skip(reason="Dependencies not implemented yet")
def test_output_exists(): 
    for sample_profile, sample_name in zip(sample_profiles, sample_names):       
        files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)

#@pytest.mark.skip(reason="Dependencies not implemented yet")
def test_all_files_are_equal():
    for sample_profile, sample_name in zip(sample_profiles, sample_names):
        files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)
    for sample in os.listdir(module_input):
        for sam in os.listdir(os.path.join(module_expected,sample)):
            if not sam.endswith('.sam'):
                continue
            files_generator.bam_to_sam(os.path.join(module_output,sample,sam.replace('.sam','.bam')), os.path.join(module_output,sample,sam))
            p, o = util.sam_to_df(os.path.join(module_expected,sample,sam)), util.sam_to_df(os.path.join(module_output,sample,sam))
            both = pd.concat([p,o], ignore_index=True).reset_index(drop=True)
            for (r, f), g in both.groupby(['QNAME','FLAG']):
                if len(g['SEQ'].iloc[0]) == 50 and g['CIGAR'].iloc[0].count('X') > 3 or \
                   len(g['SEQ'].iloc[0]) == 50 and g['CIGAR'].iloc[0].count('D') > 2 or \
                    len(g['SEQ'].iloc[0]) == 150 and g['CIGAR'].iloc[0].count('X') > 5:
                    assert len(g) == 1, 'SAM: {}: Read {} with flag {} and CIGAR {} is here and shouldn\'t be'.format(sam, r,f,g['CIGAR'].iloc[0])
                    continue
                
                assert len(g) == 2, 'SAM: {}: Read {} with flag {} is missing'.format(sam, r,f)
                for col in g.columns:
                    if col not in ['RNEXT', 'PNEXT', 'MAPQ','TLEN','CIGAR']:
                        val_pred = g[col].iloc[0]
                        val_out = g[col].iloc[1]
                        assert val_out == val_pred, 'SAM: {}: Read {} with flag {} has different {} values: \n{} \n{}'.format(sam,r,f,col, val_pred, val_out)

if __name__ == '__main__':
    test_make_files()
    test_run()
    test_output_exists()
    test_all_files_are_equal()



import click
import os
from click_option_group import optgroup

# Default values for options
OUT_DIR = os.getcwd()
SAMPLE = 'unnamed_sample'
FASTA = ""
FASTQ1 = ""
FASTQ2 = ""
DEFAULT_INTERLEAVED_INPUT = False
DEFAULT_INTERLEAVE_OUTPUT = True
INPUT_DIR = '.'
SAMPLES = None
CLUSTERING_FILE = None
#LIBRARY = 'library.csv'
LIBRARY = ""  # only read a library CSV file if told to do so
COORDS = list()
PRIMERS = list()
FILL = False
PARALLEL = 'auto'


# Common input arguments
argi_fasta = click.argument("fasta", type=click.Path(exists=True, dir_okay=False))  # path to FASTA file
argi_fastq = click.argument("fastq", type=click.Path(exists=True, dir_okay=False))  # path to FASTQ file
argi_bams = click.argument("bam_dirs", nargs=-1, type=click.Path(exists=True))  # path to one or more directories containing BAM files

# Common options
sample = optgroup.option('--sample', '-s', default=SAMPLE, type=click.Path(exists=True), help='Name of the sequence alignment map file(s) folder')
fasta = optgroup.option('--fasta', '-fa', default=FASTA, type=click.Path(exists=True), help='Path to the fasta file')
fastq1 = optgroup.option('--fastq1', '-fq1', default=FASTQ1, help='Paths to the fastq1 file. Enter multiple times for multiple files', type=click.Path(exists=True))
fastq2 = optgroup.option('--fastq2', '-fq2', default=FASTQ2, help='Paths to the fastq2 file. Enter multiple times for multiple files', type=click.Path(exists=True))
input_dir = optgroup.option('--input_dir', '-id', default=INPUT_DIR, type=click.Path(exists=True), help='Sequence alignment map files folder(s) generated by alignment')
samples = optgroup.option('--samples', '-s', default=SAMPLES, type=click.Path(exists=True), help='Path to the samples.csv file')
clustering_file = optgroup.option('--clusters', '-cl', default=CLUSTERING_FILE, type=click.Path(exists=True), help='Path to the clustering.json file')
library = optgroup.option('--library', '-l', default=LIBRARY, type=click.Path(exists=True), help='Path to the library.csv file')
interleaved = optgroup.option('--interleaved', '-i', default=DEFAULT_INTERLEAVED_INPUT, type=bool, help='Fastq files are interleaved')

opti_fastqu = click.option("--fastqu", "-u", default="", type=click.Path(exists=False), help="FASTQ file of unpaired reads")
opti_fastqi = click.option("--fastqi", "-i", default="", type=click.Path(exists=False), help="FASTQ file of interleaved paired reads")
opti_fastq1 = click.option("--fastq1", "-1", default="", type=click.Path(exists=False), help="FASTQ file of mate 1 paired reads")
opti_fastq2 = click.option("--fastq2", "-2", default="", type=click.Path(exists=False), help="FASTQ file of mate 2 paired reads")

opti_library = click.option('--library', '-l', default=LIBRARY, type=click.Path(exists=False), help='Path to a library CSV file')
opti_coords = click.option('--coords', '-c', type=(str, int, int), multiple=True, help="coordinates for reference: '-c ref-name first last'", default=COORDS)
opti_primers = click.option('--primers', '-p', type=(str, int, int), multiple=True, help="primers for reference: '-c ref-name fwd-seq rev-seq'", default=PRIMERS)
opti_fill = click.option('--fill/--no-fill', type=bool, default=FILL, help="Fill in coordinates of reference sequences for which neither coordinates nor primers were given (default: NO).")
opti_parallel = click.option('--parallel', '-P', type=click.Choice(["profiles", "reads", "off", "auto"], case_sensitive=False), default=PARALLEL, help="Parallelize the processing of mutational PROFILES or READS within each profile, turn parallelization OFF, or AUTO matically choose the parallelization method (default: AUTO).")
opto_out_dir = click.option('--out_dir', '-o', default=OUT_DIR, type=click.Path(exists=True), help=f'Where to output files (default: {OUT_DIR})')


coords = optgroup.option('--coords', '-c', type=(str, int, int), multiple=True, help="coordinates for reference: '-c ref-name first last'", default=COORDS)
primers = optgroup.option('--primers', '-p', type=(str, str, str), multiple=True, help="primers for reference: '-p ref-name fwd rev'", default=PRIMERS)
fill = optgroup.option('--fill/--no-fill', type=bool, default=FILL, help="Fill in coordinates of reference sequences for which neither coordinates nor primers were given (default: no).")
parallel = optgroup.option('--parallel', '-P', type=click.Choice(["profiles", "reads", "off", "auto"], case_sensitive=False), default=PARALLEL, help="Parallelize the processing of mutational PROFILES or READS within each profile, turn parallelization OFF, or AUTO matically choose the parallelization method (default: auto).")


# Demultiplexing
DEFAULT_DEMULTIPLEXED = False
BARCODE_START = None
BARCODE_LENGTH = None
MAX_BARCODE_MISMATCHES = 1

demultiplexing = optgroup.option('--demultiplexing', '-dx', type=bool, help='Use demultiplexing', default=DEFAULT_DEMULTIPLEXED)
barcode_start = optgroup.option('--barcode_start', '-bs', type=int, help='Start position of the barcode in the read', default=BARCODE_START)
barcode_length = optgroup.option('--barcode_length', '-bl', type=int, help='Length of the barcode', default=BARCODE_LENGTH)
max_barcode_mismatches = optgroup.option('--max_barcode_mismatches', '-mb', type=int, help='Maximum number of mutations on the barcode', default=MAX_BARCODE_MISMATCHES)


# Cutadapt parameters
DEFAULT_TRIM = True
DEFAULT_MIN_BASE_QUALITY = 25
DEFAULT_ILLUMINA_ADAPTER = "AGATCGGAAGAGC"
DEFAULT_MIN_OVERLAP = 6
DEFAULT_MAX_ERROR = 0.1
DEFAULT_INDELS = True
DEFAULT_NEXTSEQ_TRIM = True
DEFAULT_DISCARD_TRIMMED = False
DEFAULT_DISCARD_UNTRIMMED = False
DEFAULT_MIN_LENGTH = 20


# Bowtie 2 parameters
DEFAULT_LOCAL = True
DEFAULT_UNALIGNED = False
DEFAULT_DISCORDANT = False
DEFAULT_MIXED = False
DEFAULT_DOVETAIL = False
DEFAULT_CONTAIN = True
DEFAULT_FRAG_LEN_MIN = 0
DEFAULT_FRAG_LEN_MAX = 300  # maximum length of a 150 x 150 read
DEFAULT_SCORE_MIN = "L,4,0.8"
DEFAULT_N_CEILING = "L,0,0.05"
DEFAULT_SEED_INTERVAL = "L,1,0.1"
DEFAULT_GAP_BAR = 4
DEFAULT_SEED_SIZE = 20
DEFAULT_EXTENSIONS = 5
DEFAULT_RESEED = 1
DEFAULT_PADDING = 4
DEFAULT_ALIGN_THREADS = os.cpu_count()
MATCH_BONUS = "1"
MISMATCH_PENALTY = "1,1"
N_PENALTY = "0"
REF_GAP_PENALTY = "0,1"
READ_GAP_PENALTY = "0,1"
IGNORE_QUALS = True


# Clustering
CLUSTERING = False
MAX_CLUSTERS = 3
SIGNAL_THRESH = 0.005
INFO_THRESH = None
INCLUDE_G_U = False
INCLUDE_DEL = False
MIN_READS = 1000
CONVERGENCE_CUTOFF = 0.5
MIN_ITER = 100
NUM_RUNS = 10
N_CPUS = 2

clustering = optgroup.option('--clustering', '-cl', type=bool, help='Use clustering', default=CLUSTERING)
max_clusters = optgroup.option('--max_clusters', '-mc', type=int, help='Maximum number of clusters', default=MAX_CLUSTERS)
min_iter = optgroup.option('--min_iter', '-mi', type=int, help='Minimal number of EM iterations', default=MIN_ITER)
signal_thresh = optgroup.option('--signal_thresh', '-st', type=float, help='Signal threshold', default=SIGNAL_THRESH)
info_thresh = optgroup.option('--info_thresh', '-it', type=float, help='Information threshold', default=INFO_THRESH)
include_g_u = optgroup.option('--include_g_u', '-igu', type=bool, help='Include G and U', default=INCLUDE_G_U)
include_del = optgroup.option('--include_del', '-id', type=bool, help='Include deletions', default=INCLUDE_DEL)
min_reads = optgroup.option('--min_reads', '-mr', type=int, help='Minimum number of reads', default=MIN_READS)
convergence_cutoff = optgroup.option('--convergence_cutoff', '-cc', type=float, help='Convergence cutoff', default=CONVERGENCE_CUTOFF)
num_runs = optgroup.option('--num_runs', '-nr', type=int, help='Number of runs', default=NUM_RUNS)
n_cpus = optgroup.option('--n_cpus', '-cpu', type=int, help='Number of CPUs', default=N_CPUS)

# Aggregation
RNASTRUCTURE_PATH = None
RNASTRUCTURE_TEMPERATURE = False
RNASTRUCTURE_FOLD_ARGS = None
RNASTRUCTURE_DMS = False
RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE = 0.04
RNASTRUCTURE_DMS_MAX_PAIRED_VALUE = 0.01
RNASTRUCTURE_PARTITION = False
RNASTRUCTURE_PROBABILITY = False
POISSON = True

rnastructure_path = optgroup.option('--rnastructure_path', '-rs', type=click.Path(exists=True), help='Path to RNAstructure, to predict structure and free energy', default=RNASTRUCTURE_PATH)
rnastructure_temperature = optgroup.option('--rnastructure_temperature', '-rst', type=bool, help='Use sample.csv temperature values for RNAstructure', default=RNASTRUCTURE_TEMPERATURE)
rnastructure_fold_args = optgroup.option('--rnastructure_fold_args', '-rsa', type=str, help='optgroup.options to pass to RNAstructure fold', default=RNASTRUCTURE_FOLD_ARGS)
rnastructure_dms = optgroup.option('--rnastructure_dms', '-rsd', type=bool, help='Use the DMS signal to make predictions with RNAstructure', default=   RNASTRUCTURE_DMS)
rnastructure_dms_min_unpaired_value = optgroup.option('--rnastructure_dms_min_unpaired_value', '-rsdmin', type=int, help='Minimum unpaired value for using the dms signal as an input for RNAstructure', default=RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE)
rnastructure_dms_max_paired_value = optgroup.option('--rnastructure_dms_max_paired_value', '-rsdmax', type=int, help='Maximum paired value for using the dms signal as an input for RNAstructure', default=RNASTRUCTURE_DMS_MAX_PAIRED_VALUE)
rnastructure_partition = optgroup.option('--rnastructure_partition', '-rspa', type=bool, help='Use RNAstructure partition function to predict free energy', default=RNASTRUCTURE_PARTITION)
rnastructure_probability = optgroup.option('--rnastructure_probability', '-rspr', type=bool, help='Use RNAstructure partition function to predict per-base mutation probability', default=RNASTRUCTURE_PROBABILITY)
poisson = optgroup.option('--poisson', '-po', type=bool, help='Predict Poisson confidence intervals', default=POISSON)

# Misc
VERBOSE = False
verbose = optgroup.option('--verbose', '-v', type=bool, help='Verbose output', default=VERBOSE)






import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import pandas as pd
import dreem.util as util
from dreem import vectoring, aggregation
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
from dreem.test.remove_output_test import remove_module_files
import pytest
import json
from dreem.util.files_sanity import compare_fields
from dreem.test.sample_profile import sample_profile, constructs, sections


module = 'vectoring_aggregate'
sample_name = 'test_set_1'

module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

inputs = ['bam','fasta','library','samples']
outputs = ['output']

#@pytest.mark.skip(reason="no way of currently testing this")
def test_make_files():
    remove_module_files(module)
    files_generator.generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name, rnastructure_config={'path':'/Users/ymdt/src/RNAstructure/exe', 'temperature':False, 'fold_args':'', 'dms':False, 'dms_min_unpaired_value':0.01, 'dms_max_paired_value':0.06, 'partition':False, 'probability':False })
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, outputs, prediction_dir, sample_name)

    
def test_run():
    for sample in os.listdir(module_input):
        vectoring.run(
            bam_dirs = [os.path.join(module_input, sample, f) for f in os.listdir(os.path.join(module_input, sample)) if f.endswith('.bam')],
            out_dir = os.path.join(module_input,sample),
            fasta = os.path.join(module_input, sample, 'reference.fasta'),
            library = os.path.join(module_input, sample, 'library.csv'),
            )
        aggregation.run(
            bv_files= [os.path.join(module_input, sample,'vectoring', sample, c, s) for c in constructs for s in sections[constructs.index(c)]],
            out_dir = module_output,
            samples = os.path.join(module_input, sample_name, 'samples.csv'),
            library= os.path.join(module_input, sample_name, 'library.csv'),
            sample=sample,
            fasta = os.path.join(module_input, sample_name, 'reference.fasta'),
            rnastructure_path='/Users/ymdt/src/RNAstructure/exe',
            poisson= True
            #clusters = os.path.join(module_input, sample, 'clustering.csv') #TODO
        )
    global output, expected
    expected = json.load(open(os.path.join(module_expected, sample_name+'.json'), 'r'))
    output = json.load(open(os.path.join(module_output, sample_name+'.json'), 'r'))

def test_output_exists():        
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)

@pytest.mark.parametrize('attr',['date', 'sample','user','temperature_k','inc_time_tot_secs','exp_env','DMS_conc_mM'])
def test_sample_attributes(attr):
    compare_fields(expected, output, [attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr',['sequence', 'barcode', 'barcode_start', 'some_random_attribute'])
def test_library_attributes(construct,attr):
    compare_fields(expected, output, [construct, attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr',['num_reads', 'num_aligned','skips_short_reads','skips_too_many_muts','skips_low_mapq'])
def test_alignment_attributes(construct,attr):
    compare_fields(expected, output, [construct, attr])

@pytest.mark.parametrize('construct,section', [(c,s) for c in constructs for s in sections[constructs.index(c)]])
@pytest.mark.parametrize('attr',['section_start','section_end','sequence','structure','deltaG'])
def test_sections_attributes(construct,section,attr):
    compare_fields(expected, output, [construct, section, attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr', ['worst_cov_bases','mut_bases', 'del_bases', 'ins_bases', 'cov_bases', 'info_bases', 'mut_rates', 'mod_bases_A', 'mod_bases_C', 'mod_bases_G', 'mod_bases_T','poisson_high','poisson_low'] ) #'num_of_mutations'
def test_mp_pop_avg(construct,attr):
    for section in sections[constructs.index(construct)]:
        compare_fields(expected, output, [construct, section, 'pop_avg', attr])

@pytest.mark.parametrize('construct', constructs)
def test_section_idx(construct):
    for section in sections[constructs.index(construct)]:
        ss, se = output[construct][section]['section_start'], output[construct][section]['section_end']
        assert len(output[construct][section]['sequence']) == se-ss+1, 'section length is not correct: {} != {}-{}+1'.format(len(output[construct][section]['sequence']), se, ss)
        assert output[construct]['sequence'].index(output[construct][section]['sequence']) == ss-1, 'section start is not correct'
        assert output[construct]['sequence'][ss-1:se] == output[construct][section]['sequence'], 'section sequence is not correct'
        

if __name__ == '__main__':
    # remove all files
    test_make_files()
    test_run()



from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import subprocess, os

def test_remove_files():
    os.system(' '.join(['rm', '-fr', test_files_dir]))
    
def remove_module_files(module):
    for folder in ['input', 'output', 'expected_output']:
        os.system('rm -rf {}'.format(os.path.join(test_files_dir, folder, module)))
        os.makedirs(os.path.join(test_files_dir, folder, module), exist_ok=True)



import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import dreem, os
import dreem.util as util
import pandas as pd
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import pytest
import dreem.pipeline
import shutil
from dreem.util.files_sanity import compare_fields
import json
from dreem.test.sample_profile import sample_profile, constructs, sections

sample_name = 'test_set_1'
module = 'pipeline'


inputs = ['fastq','fasta','samples','library']
outputs = ['output']

module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

# ### Create test files for `test set 1`
def test_make_files():
    for folder in ['input', 'output', 'expected_output']:
        os.system('rm -rf {}'.format(os.path.join(test_files_dir, folder, module)))
        os.makedirs(os.path.join(test_files_dir, folder, module), exist_ok=True)
    files_generator.generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name, rnastructure_config={'path':'/Users/ymdt/src/RNAstructure/exe', 'temperature':False, 'fold_args':'', 'dms':False, 'dms_min_unpaired_value':0.01, 'dms_max_paired_value':0.06, 'partition':False, 'probability':False })
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, outputs, prediction_dir, sample_name)
    global expected
    expected = json.load(open(os.path.join(module_expected, sample_name+'.json'), 'r'))

def test_run():
    for sample in os.listdir(module_input):
        
        dreem.pipeline.run(            
            fastq1 = '{}/{}_R1.fastq'.format(os.path.join(module_input,sample),sample),\
            fastq2 = '{}/{}_R2.fastq'.format(os.path.join(module_input,sample),sample),\
            fasta = '{}/reference.fasta'.format(os.path.join(module_input,sample)),\
            library = '{}/library.csv'.format(os.path.join(module_input,sample)),\
            samples = '{}/samples.csv'.format(os.path.join(module_input,sample)),\
            out_dir = module_output,
            rnastructure_path='/Users/ymdt/src/RNAstructure/exe'
            )
    global output
    output = json.load(open(os.path.join(module_output, 'output','aggregation',sample_name+'.json'), 'r'))
        
def test_output_exists():        
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)

def test_output_exists():        
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)

@pytest.mark.parametrize('attr',['date', 'sample','user','temperature_k','inc_time_tot_secs','exp_env','DMS_conc_mM'])
def test_sample_attributes(attr):
    compare_fields(expected, output, [attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr',['sequence', 'barcode', 'barcode_start', 'some_random_attribute'])
def test_library_attributes(construct,attr):
    compare_fields(expected, output, [construct, attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr',['num_reads', 'num_aligned','skips_short_reads','skips_too_many_muts','skips_low_mapq'])
def test_alignment_attributes(construct,attr):
    compare_fields(expected, output, [construct, attr])

@pytest.mark.parametrize('construct,section', [(c,s) for c in constructs for s in sections[constructs.index(c)]])
@pytest.mark.parametrize('attr',['section_start','section_end','sequence','structure','deltaG'])
def test_sections_attributes(construct,section,attr):
    compare_fields(expected, output, [construct, section, attr])

@pytest.mark.parametrize('construct', constructs)
@pytest.mark.parametrize('attr', ['worst_cov_bases','mut_bases', 'del_bases', 'ins_bases', 'cov_bases', 'info_bases', 'mut_rates', 'mod_bases_A', 'mod_bases_C', 'mod_bases_G', 'mod_bases_T','poisson_high','poisson_low'] ) #'num_of_mutations'
def test_mp_pop_avg(construct,attr):
    for section in sections[constructs.index(construct)]:
        compare_fields(expected, output, [construct, section, 'pop_avg', attr])

@pytest.mark.parametrize('construct', constructs)
def test_section_idx(construct):
    for section in sections[constructs.index(construct)]:
        ss, se = output[construct][section]['section_start'], output[construct][section]['section_end']
        assert len(output[construct][section]['sequence']) == se-ss+1, 'section length is not correct: {} != {}-{}+1'.format(len(output[construct][section]['sequence']), se, ss)
        assert output[construct]['sequence'].index(output[construct][section]['sequence']) == ss-1, 'section start is not correct'
        assert output[construct]['sequence'][ss-1:se] == output[construct][section]['sequence'], 'section sequence is not correct'
        
if __name__ == '__main__':
    # remove test files
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'output', module)))
    test_make_files()
    test_run()



import numpy as np
import datetime
from os.path import exists
import os
import matplotlib.pyplot as plt
from typing import Tuple, List
import shutil
from plotly.validators.scatter.marker import SymbolValidator
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPM
from scipy.optimize import curve_fit
import scipy

vals = SymbolValidator().values

def Setshape(x):
        vals = SymbolValidator().values
        return vals[3*x]


def make_path(path:str)->str:
    """Create directories until path exists on your computer. Turns the keyword 'date' into today's date.

    Args:
        path: series of directories that you want to create.
    
    Returns:
        Updated path with today's date instead of the keyword 'date'  
    """

    path = os.path.normpath(path)
    path=path.split(os.sep)
    try:
        path[path.index('date')] = str(datetime.datetime.now())[:10]
    except:
        'No date in path'
    full_path = ''
    for repo in path:
        full_path = full_path + f"{repo}/"
        if not exists(full_path):
            os.mkdir(full_path)
    return full_path



def gini(x:np.array)->float:
    """Returns Gini index

    Args:
        x (np.array): the array you want the Gini index from

    Returns:
        float: Gini index of the input array
    """
    # (Warning: This is a concise implementation, but it is O(n**2)
    # in time and memory, where n = len(x).  *Don't* pass in huge
    # samples!)

    # Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
    # Relative mean absolute difference
    rmad = mad/np.mean(x)
    # Gini coefficient
    g = 0.5 * rmad
    return g

def savefig(file:str, close=True)->None:
    """Save a matplotlib figure and create the directory if it doesn't exists.

    Args:
        file: path+title.
        facecolor: color of the background 
    """

    path = make_path('/'.join(file.split('/')[:-1]))
    plt.savefig(path+file.split('/')[-1], bbox_inches='tight')
    if close:
        # Clear the current axes.
        plt.cla() 
        # Clear the current figure.
        plt.clf() 
        # Closes all the figure windows.
        plt.close('all')   


def define_figure(title:str, xlabel:str, ylabel:str, figsize:Tuple[float, float])->plt.figure:
    """Define title, labels and size of your figure.

    Args:
        title: matplotlib title
        xlabel: matplotlib xlabel
        ylabel: matplotlib ylabel
        figsize: matplotlib figsize
    """

    fig = plt.figure(figsize=figsize)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    return fig

import yaml, os, subprocess
import pandas as pd


class Fit(object):
    def __init__(self) -> None:
        self.legend = ''
    
    def get_legend(self):
        return self.legend

    def predict(self, x, y, model, prefix='', suffix=''):
        fit = self.fit(x,y,model)
        m = eval(model)
        try:
            linreg  = scipy.stats.linregress(y,m(x,*fit))
            self.rvalue = round(linreg.rvalue,5)
        except:
            self.rvalue = 'error'

        self._generate_legend(fit, model, prefix, suffix)
        return np.sort(x), m(np.sort(x),*fit)

    def fit(self, x,y, model):
        fit = curve_fit(eval(model), x, y)[0]
        return fit

    def _generate_legend(self, fit, m, prefix, suffix):
        slice_m = lambda start, stop: ','.join(str(m).split(',')[start:stop])
        first_slice = slice_m(0,len(fit))+','+slice_m(len(fit), len(fit)+1).split(':')[0]
        second_slice = ','.join(m.split(',')[len(fit):])[2:]
        fun_args = [a.strip() for a in str(m).split(',')[1:len(fit)+1]]
        fun_args[-1] = fun_args[-1][0]
        for a,b in zip(fun_args,fit):
            second_slice = second_slice.replace(a.strip(),str(round(b,5)))
        self.legend = prefix+ second_slice + suffix +f'\n R2={self.rvalue}'

                
class RNAstructure(object): 
    """A class to run RNAstructure commands

    Args:
        object (_type_): _description_

    Examples:
        rna = RNAstructure('/Users/ymdt/src/RNAstructure/exe')
        rna.fit(sequence='ACCTTTCAGAGCTACGATCGACTAGCTAGCATCGATACAGCGACACAAGCATTTGTAGCATTAGGTCA')
        print("DeltaG + structure:", rna.predict_construct_deltaG())
        print("Ensemble energy:", rna.predict_ensemble_energy())
        print("Partition function:", rna.predict_partition()) 
    """
    def __init__(self, rnastructure_path) -> None:
        self.rnastructure_path = rnastructure_path if rnastructure_path[-1] == '/' else rnastructure_path+'/'

    def fit(self, sequence, construct='construct'):
        self.sequence = sequence
        temp_folder = 'temp/rnastructure/'
        self.__make_temp_folder(temp_folder)
        self.__make_files(temp_folder+'temp')
        self.__create_fasta_file(construct, sequence)

    def predict_ensemble_energy(self):
        cmd = f"{self.rnastructure_path}EnsembleEnergy {self.fasta_file} --DNA --sequence"
        splitted_output = self.__run_command(cmd)[0].split(' ')
        return float(splitted_output[splitted_output.index(f"kcal/mol\n\nEnsemble")-1])

    def predict_partition(self, temperature_k =None):
        cmd = f"{self.rnastructure_path}partition {self.fasta_file} {self.pfs_file} --DNA"
        if temperature_k != None:
            cmd += ' --temperature '+str(temperature_k)
        self.__run_command(cmd)
        self.__run_command(self.rnastructure_path+'ProbabilityPlot '+ self.pfs_file + ' -t '+self.prob_file)
        with open(self.prob_file,"r") as f:
            lines=f.readlines()
            out={'i':[],'j':[],'p':[]}
            for x in range(len(lines)):
                if x>1:
                    ls=lines[x].split("\t")
                    out["i"]+=[int(ls[0])]
                    out["j"]+=[int(ls[1])]
                    out["p"]+=[float(ls[2])]
        return self.__cast_pairing_prob(out)

    def draw(self, savefig, mut_rates = [], dpi=72):
        self.predict_construct_deltaG()
        if len(mut_rates):
            mut_rates = np.array(mut_rates).reshape(-1,1)
            with open(self.s_file, 'w') as f:
                for i in range(len(mut_rates)):
                    f.write(f'{i+1} {mut_rates[i][0] if mut_rates[i][0] != np.nan else -999 } \n')
                f.close()
            
        cmd = self.rnastructure_path+'draw '+self.ct_file + ' --svg '+self.svg_file+' -n -1'
        self.__run_command(cmd if len(mut_rates) == 0 else cmd+' -s '+self.s_file)
        svg_code = self.__read_svg()
        drawing = svg2rlg(self.svg_file)
        renderPM.drawToFile(drawing, self.png_file, fmt="PNG", dpi=dpi)
        shutil.copy(self.png_file, savefig)

    def __read_svg(self):
        with open(self.svg_file, "r") as f:
            svg_code = f.read()
        return svg_code

    def predict_construct_deltaG(self, temperature_k=None):
        # Run RNAstructure
        suffix = ''
        cmd = f"{self.rnastructure_path}Fold {self.fasta_file} {self.ct_file} -d" + suffix
        self.__run_command(cmd)
        assert os.path.getsize(self.ct_file) != 0, f"{self.ct_file} is empty, check that RNAstructure works"
        self.deltaG, self.structure =  self.__extract_deltaG_struct()
        return self.deltaG, self.structure

    def __make_temp_folder(self, temp_folder):
        isExist = os.path.exists(temp_folder)
        if not isExist:
            os.makedirs(temp_folder)
        return temp_folder

    def __make_files(self, temp_prefix='temp/temp'):
        self.pfs_file = f"{temp_prefix}.pfs"
        self.ct_file = f"{temp_prefix}.ct"
        self.dot_file = f"{temp_prefix}_dot.txt"
        self.fasta_file = temp_prefix+'.fasta'
        self.prob_file = temp_prefix+'_prob.txt'
        self.svg_file = temp_prefix+'.svg'
        self.png_file = temp_prefix+'.png'
        self.s_file = temp_prefix+'_s.txt'

    def __create_fasta_file(self, construct, sequence):
        # push the ref into a temp file
        temp_fasta = open(self.fasta_file, 'w')
        temp_fasta.write('>'+construct+'\n'+sequence)
        temp_fasta.close()

    # cast the temp file into a dot_bracket structure and extract the attributes
    def __extract_deltaG_struct(self):
        self.__run_command(f"ct2dot {self.ct_file} 1 {self.dot_file}")
        temp_dot = open(self.dot_file, 'r')
        first_line = temp_dot.readline().split()
        # If only dots in the structure, no deltaG 
        out = {}
        if len(first_line) == 4:
            _, _, deltaG, _ = first_line
            deltaG = float(deltaG)
        if len(first_line) == 1:
            deltaG, _ = 'void', first_line[0][1:]

        sequence = temp_dot.readline()[:-1] #  Remove the \n
        structure = temp_dot.readline()[:-1] # Remove the \n
        return deltaG, structure

    def __cast_pairing_prob(self, prob:dict)->list:
        """Computes the pairing probability list.
        Args:
            prob (dict): Ouput of RNAstructure.
            index (int): index of the row that you want the pairing probabilities of.
        Returns:
            list: pairing probability, under the form of a list of probabilities
        """
        # Create local dataframe to play here
        df_loc = pd.DataFrame(prob)
        df_loc = pd.concat((df_loc, df_loc.rename(columns={'i':'j','j':'i'})))
        # Group the probabilities by i and get an ordered list of the pairing probability
        g = df_loc.groupby('i')['p'].agg(lambda row: [pow(10,-float(r)) for r in row])
        g.index = g.index.astype(int)
        g = g.sort_index()
        g['sum_log_p'] = g.apply(lambda row: sum(row))
        return list(g['sum_log_p'])

    def __run_command(self, cmd):
        output, error_msg = None, None
        try:
            output = subprocess.check_output(
                    cmd, shell=True, stderr=subprocess.STDOUT
            ).decode("utf8")
        except subprocess.CalledProcessError as exc:
            error_msg = exc.output.decode("utf8")
        return output, error_msg




import numpy as np
import json

def cast_to_json_compat(obj):
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    return obj

def cast_dict(mp):
    for k, v in mp.items():
        mp[k] = cast_to_json_compat(v)
        if type(v) is dict:
            cast_dict(v)
    return mp
    
def sort_dict(mut_profiles):
    sorting_key = lambda item: 1000*{int:0, str:0, float:0, list:1, dict:2}[type(item[1])] + ord(item[0][0])
    mut_profiles = {k:mut_profiles[k] for k in sorted(mut_profiles)}
    mut_profiles = dict(sorted(mut_profiles.items(), key=sorting_key))
    for k, v in mut_profiles.items():
        if type(v) is not dict:
            continue
        mut_profiles[k] = sort_dict(v)
    return mut_profiles


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)
    
def flatten_json(data):
    out, row = [], {}
        
    for k,v in data.copy().items():
        if type(v) != dict:
            row[k]= v
        else:
            row['construct'] = k
            for k2,v2 in v.items():
                if type(v2) != dict:
                    row[k2] = v2
                else:
                    row['section'] = k2
                    for k3,v3 in v2.items():
                        row['cluster'] = k3
                        if type(v3) != dict:
                            row[k3] = v3
                        else:
                            for k4,v4 in v3.items():
                                row[k4] = v4
                            out.append(row.copy())
    return out


from __future__ import annotations
from dataclasses import dataclass, asdict, astuple
import itertools
import os
from pathlib import Path
import re
from typing import Callable, Dict, Tuple, Optional, Any


# Partitions
OUTPUT_DIR = "output"
TEMP_DIR = "temp"
PARTITIONS = (OUTPUT_DIR, TEMP_DIR)

# Module directories
MOD_DMX = "demultiplexing"
MOD_ALN = "alignment"
MOD_VEC = "vectoring"
MOD_CLS = "clustering"
MOD_AGG = "aggregation"
MODULES = (MOD_DMX, MOD_ALN, MOD_VEC, MOD_CLS, MOD_AGG)

# Alignment steps
ALN_TRIM = "align_1_trim"
ALN_ALIGN = "align_2_align"
ALN_REM = "align_3_rem"
ALN_SORT = "align_4_sort"
ALN_SPLIT = "align_5_split"
ALN_STEPS = (ALN_TRIM, ALN_ALIGN, ALN_REM, ALN_SORT, ALN_SPLIT)

# Vectoring steps
VEC_SELECT = "vector_1_select"
VEC_SORT = "vector_2_sort"
VEC_STEPS = (VEC_SELECT, VEC_SORT)

TEMP_STEPS = ALN_STEPS + VEC_STEPS

# File extensions
FASTA_EXTS = (".fasta", ".fa")
FQ_EXTS = (".fastq", ".fq", ".fastq.gz", ".fq.gz")
FQ_PATTERNS = ("_r{}{}", "_mate{}{}", "_{}_sequence{}")
FQ_CASES = (str.lower, str.upper)
FQ1_EXTS = tuple(case(ptrn).format(1, ext) for case, ptrn, ext in
                 itertools.product(FQ_CASES, FQ_PATTERNS, FQ_EXTS))
FQ2_EXTS = tuple(case(ptrn).format(2, ext) for case, ptrn, ext in
                 itertools.product(FQ_CASES, FQ_PATTERNS, FQ_EXTS))
SAM_EXT = ".sam"
BAM_EXT = ".bam"
CRAM_EXT = ".cram"
XAM_EXTS = (SAM_EXT, BAM_EXT, CRAM_EXT)
XAI_EXTS = (f"{BAM_EXT}.bai",)


# Path functions

def sanitize(path: Any):
    return os.path.realpath(os.path.normpath(os.path.abspath(str(path))))


def switch_directory(old_path: str, new_dir: str):
    return os.path.join(new_dir, os.path.basename(old_path))


def try_remove(file: str):
    try:
        os.remove(file)
    except OSError:
        pass


# 

class PathError(BaseException):
    pass


# Path segment classes


class PathSegment(object):
    def __init__(self, segment: Any) -> None:
        self._assert_valid(segment)
        self._segment = str(segment)
    
    @classmethod
    def _assert_valid(cls, segment: Any):
        if not cls.segment_is_valid(segment):
            raise PathError(f"Invalid segment for {cls}: '{segment}'")
    
    def __str__(self) -> str:
        return self._segment
    
    def __repr__(self) -> str:
        return f"'{self}'"
    
    @staticmethod
    def segment_is_valid(segment: Any):
        return True


class BaseSegment(PathSegment):
    def __init__(self, segment: Any) -> None:
        super().__init__(sanitize(segment))
    
    segment_is_valid = staticmethod(lambda seg: os.path.isdir(str(seg)))
    
    @property
    def path(self):
        return Path(self._segment)


class SubSegment(PathSegment):
    segment_is_valid = staticmethod(lambda seg: os.sep not in str(seg))
    
    @property
    def name(self):
        return self._segment


class PatternedSegment(SubSegment):
    _pattern: str = "(.+)"
    _dtypes: Tuple[type] = (str,)
    _exts: Tuple[str] = ()
    _fmt = staticmethod(lambda ext: "{}" + ext)

    @classmethod
    def _get_patterns(cls):
        return (re.compile(f"^{cls._pattern}({ext.replace('.', '[.]')})$")
                for ext in cls._exts)

    @classmethod
    def _cast_parsed_groups(cls, parsed_groups: tuple) -> tuple:
        return tuple(dtype(group) for dtype, group
                     in zip(cls._dtypes, parsed_groups, strict=True))
    
    @classmethod
    def _clean_ext(cls, ext: str) -> str:
        if not cls._exts:
            raise PathError(f"No extensions for {cls}")
        if not ext:
            ext = cls._exts[0]
        elif ext not in cls._exts:
            raise PathError(f"Invalid extension for {cls}: '{ext}'")
        return ext
    
    @classmethod
    def _format_noparse(cls, *args, ext: str) -> str:
        segment = cls._fmt(cls._clean_ext(ext)).format(*args)
        return segment
    
    @classmethod
    def _parse_noformat(cls, segment: Any) -> tuple:
        for pattern in cls._get_patterns():
            if (match := pattern.match(str(segment))):
                fields = cls._cast_parsed_groups(match.groups()[:-1])
                ext = match.groups()[-1]
                return fields, ext
        raise PathError(f"Failed to parse segment '{segment}' for class {cls}")
    
    @classmethod
    def format(cls, *args, ext: str = "") -> str:
        segment = cls._format_noparse(*args, ext=ext)
        assert cls._parse_noformat(segment) == (args, cls._clean_ext(ext))
        return segment
    
    @classmethod
    def parse(cls, segment: Any) -> tuple:
        fields, ext = cls._parse_noformat(segment)
        assert str(segment) == cls._format_noparse(*fields,
                                                   ext=cls._clean_ext(ext))
        return fields

    @classmethod
    def segment_is_valid(cls, segment: Any):
        if not super().segment_is_valid(segment):
            return False
        try:
            cls.parse(segment)
        except PathError:
            return False
        else:
            return True
    
    @property
    def fields(self):
        return self.parse(self._segment)
    
    @property
    def name(self):
        return self._fmt("").format(*self.fields)


class FastaSegment(PatternedSegment):
    _exts = FASTA_EXTS


class FastqSegment(PatternedSegment):
    _exts = FQ_EXTS


class Fastq1Segment(FastqSegment):
    _exts = FQ1_EXTS


class Fastq2Segment(FastqSegment):
    _exts = FQ2_EXTS


class XamSegment(PatternedSegment):
    _exts = XAM_EXTS


class XamIndexSegment(PatternedSegment):
    _exts = XAI_EXTS


class PartitionSegment(SubSegment):
    segment_is_valid = staticmethod(lambda seg: seg in PARTITIONS)


class ModuleSegment(SubSegment):
    segment_is_valid = staticmethod(lambda seg: seg in MODULES)


class TempStepSegment(SubSegment):
    segment_is_valid = staticmethod(lambda seg: seg in TEMP_STEPS)


class SampleSegment(SubSegment):
    pass


class RefSegment(SubSegment):
    pass


class RegionSegment(PatternedSegment):
    _pattern = "([0-9]+)-([0-9]+)"
    _dtypes = (int, int)
    _exts = ("",)
    _fmt = staticmethod(lambda ext: "{}-{}")


class MutVectorReportSegment(PatternedSegment):
    _pattern = "([0-9]+)-([0-9]+)_report"
    _dtypes = (int, int)
    _exts = (".txt",)
    _fmt = staticmethod(lambda ext: "{}-{}_report" + ext)


class MutVectorBatchSegment(PatternedSegment):
    _pattern = "vectors_([0-9]+)"
    _dtypes = (int,)
    _exts = (".orc",)
    _fmt = staticmethod(lambda ext: "vectors_{}" + ext)


# Segment descriptor class

class SegmentDescriptor(object):
    _priv_chr = "_"

    segment_types: Dict[str, type] = {
        "base": BaseSegment,
        "fasta": FastaSegment,
        "fastq": FastqSegment,
        "fastq1": Fastq1Segment,
        "fastq2": Fastq2Segment,
        "partition": PartitionSegment,
        "module": ModuleSegment,
        "temp_step": TempStepSegment,
        "sample": SampleSegment,
        "ref": RefSegment,
        "region": RegionSegment,
        "xam": XamSegment,
        "xai": XamIndexSegment,
        "mv_report": MutVectorReportSegment,
        "mv_batch": MutVectorBatchSegment,
    }

    def __set_name__(self, owner, name: str):
        if name.startswith(self._priv_chr):
            raise ValueError(f"name '{name}' starts with '{self._priv_chr}'")
        self.public_name = name
        self.private_name = f"{self._priv_chr}{name}"
        self.segment_type = self.segment_types[self.public_name]
    
    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)
    
    def __set__(self, obj, value):
        setattr(obj, self.private_name, self.segment_type(value))


# Paths

@dataclass
class BasePath(object):
    base: SegmentDescriptor = SegmentDescriptor()

    @classmethod
    def parse(cls, path: Any):
        if cls is BasePath:
            return cls(path)
        head, tail = os.path.split(str(path))
        parent, = cls.__bases__
        segments = parent.parse(head).args + (tail,)
        return cls(*segments)
    
    @property
    def segments(self):
        return astuple(self)
    
    @property
    def named_segments(self):
        return asdict(self)
    
    @property
    def args(self):
        return tuple(map(str, self.segments))
    
    @property
    def kwargs(self):
        return {name: str(seg) for name, seg in self.named_segments.items()}
    
    @property
    def path(self):
        return Path(*self.args)
    
    @property
    def last(self):
        """ Get the last segment of the path. """
        return astuple(self)[-1]
    
    def __str__(self) -> str:
        return str(self.path)


@dataclass
class PartitionPath(BasePath):
    partition: SegmentDescriptor = SegmentDescriptor()


@dataclass
class ModulePath(PartitionPath):
    module: SegmentDescriptor = SegmentDescriptor()


@dataclass
class SampleOutPath(ModulePath):
    sample: SegmentDescriptor = SegmentDescriptor()


@dataclass
class SampleInPath(BasePath):
    sample: SegmentDescriptor = SegmentDescriptor()


@dataclass
class TempStepPath(ModulePath):
    temp_step: SegmentDescriptor = SegmentDescriptor()


@dataclass
class SampleTempPath(TempStepPath):
    sample: SegmentDescriptor = SegmentDescriptor()


# Alignment Paths


@dataclass
class FastaInPath(BasePath):
    fasta: SegmentDescriptor = SegmentDescriptor()


@dataclass
class FastaSampleOutPath(SampleOutPath):
    fasta: SegmentDescriptor = SegmentDescriptor()


@dataclass
class FastqInPath(BasePath):
    fastq: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq1InPath(BasePath):
    fastq1: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq2InPath(BasePath):
    fastq2: SegmentDescriptor = SegmentDescriptor()


@dataclass
class FastqDemultiPath(SampleOutPath):
    fastq: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq1DemultiPath(SampleOutPath):
    fastq1: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq2DemultiPath(SampleOutPath):
    fastq2: SegmentDescriptor = SegmentDescriptor()


@dataclass
class FastqOutPath(SampleTempPath):
    fastq: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq1OutPath(SampleTempPath):
    fastq1: SegmentDescriptor = SegmentDescriptor()


@dataclass
class Fastq2OutPath(SampleTempPath):
    fastq2: SegmentDescriptor = SegmentDescriptor()


@dataclass
class XamTempPath(SampleTempPath):
    xam: SegmentDescriptor = SegmentDescriptor()


@dataclass
class XamOutPath(SampleOutPath):
    xam: SegmentDescriptor = SegmentDescriptor()


@dataclass
class XamIndexTempPath(SampleTempPath):
    xai: SegmentDescriptor = SegmentDescriptor()


# Vectoring Paths

@dataclass
class XamInPath(SampleInPath):
    xam: SegmentDescriptor = SegmentDescriptor()


@dataclass
class XamIndexInPath(SampleInPath):
    xai: SegmentDescriptor = SegmentDescriptor()


@dataclass
class RefPath(SampleOutPath):
    ref: SegmentDescriptor = SegmentDescriptor()


@dataclass
class RegionPath(RefPath):
    region: SegmentDescriptor = SegmentDescriptor()


@dataclass
class MutVectorReportPath(RefPath):
    mv_report: SegmentDescriptor = SegmentDescriptor()


@dataclass
class MutVectorBatchPath(RegionPath):
    mv_batch: SegmentDescriptor = SegmentDescriptor()



import pandas as pd
from tqdm import tqdm
import numpy as np
import os

def add_library_info(df, df_library, verbose=False):

    if verbose: 
        print(f"Adding library info")

    # check the sanity of libraries.csv
    assert 'construct' in list(df_library.columns), "construct is not in library.csv"

    df_library.construct = df_library.construct.astype(str)
    if not sum([c in df_library.columns.tolist() for c in ['section_start','section_end']]) == 2:
        df = pd.merge(df, df_library, on='construct', how='left')
    else:
        new_df = pd.DataFrame()
        assert len(df['construct'].unique()) == len(df), 'constructs are not unique'
        if verbose:
            iter_fun = lambda x: tqdm(x.groupby('construct'),  total=len(df_library['construct'].unique()),desc='constructs')
        else:
            iter_fun = lambda x: x.groupby('construct')
        for idx, g in iter_fun(df_library):
            one_full_construct = False
            for _, row in g.iterrows():
                if row['construct'] not in df['construct'].tolist():
                    print(f'construct {row["construct"]} not in df')
                    continue
                row = pd.concat([row, df[df['construct']==row['construct']][[c for c in df.columns if c not in row.index]].iloc[0]])
                if 'Unnamed: 0' in row.index:
                    row = row.drop('Unnamed: 0')
                if not 'section' in df_library.columns:
                    row['section'] = row['section_start'] + '-' + row['section_end']
                row.rename({'section': 'section'}, inplace=True)
                isnan = lambda x: np.isnan(x) if isinstance(x, float) else False
                if isnan(row['section_start']) or isnan(row['section_end']):
                    if not one_full_construct:
                        one_full_construct = True
                        row['section'] = 'full'
                        row['section_start'], row['section_end'] = 0, len(row['sequence'])
                else:
                    row['section_start'], row['section_end'] = int(row['section_start']), int(row['section_end']) # CAST TO INT
                    for c in ['sequence']+[col for col in df.columns.tolist() if (col != 'num_of_mutations' and type(df[col].iloc[0]) == list)]:
                        row[c] = row[c][int(row['section_start']):int(row['section_end'])]
                        if isnan(row['section']):
                            row['section'] = '{}-{}'.format(row['section_start'], row['section_end'])
                new_df = pd.concat([new_df, pd.DataFrame(row).T])
        df = new_df.reset_index(drop=True)
        return df
